/***************************************************************************
   Module Name:  cb      sdk_src_cb_cb.c

   Description:  Circular Buffer 

   Author:  Tools and Technologies Group
            Autodesk, Inc.
   
     (C) Copyright 1990, 1991 by Autodesk, Inc.
***************************************************************************
*  Permission to use, copy, modify, and distribute this software and its  *
*  documentation for the purpose of creating applications for AutoCAD, is *
*  hereby granted in accordance with the terms of the License Agreement   *
*  accompanying this product.                                             *
*                                                                         *
***************************************************************************
*                                                                         *
*  AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS.  AUTODESK  *
*  SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF MERCHANTABILITY OR      *
*  FITNESS FOR A PARTICULAR USE.  AUTODESK, INC. DOES NOT WARRANT THAT    *
*  THE OPERATION OF THE PROGRAM WILL BE UNINTERRUPTED OR ERROR FREE.      *
*                                                                         *
***************************************************************************
*                                                                         *
*                       RESTRICTED RIGHTS LEGEND                          *
*                                                                         *
*   Use, duplication, or disclosure by the U.S. Government is subject     *
*   to restrictions set forth in FAR 52.227-19 (Commercial Computer       *
*   Software - Restricted Rights) and DFAR 252.227-7013 (c) (1) (ii)      *
*   (Rights in Technical Data and Computer Software), as applicable.      *
***************************************************************************
*                                                                         *
*   Last Modification:                                                    *
*                                                                         *
*   Notes and restrictions on use:                                        *
***************************************************************************/
/**************************************************************************/
/*  STANDARD C INCLUDE FILES  */
/**************************************************************************/
#include  <stdio.h>
#include  "sdk.h"


/**************************************************************************/
/*  GLOBAL DEFINES  */
/**************************************************************************/


/**************************************************************************/
/*  GLOBAL VARIABLES  */
/**************************************************************************/
void         **cb = (void **)NULL;  /* beginning of buffer */
void         **cp = (void **)NULL;  /* current pointer */
void         **ep = (void **)NULL;  /* end pointer */
void         **rp = (void **)NULL;  /* read pointer */


/**************************************************************************/
/*  GLOBAL FUNCTIONS  */
/**************************************************************************/
void         **cb_alloc_node();     /* node allocation function */
int          cb_write();            /* node write function */
void         *cb_read();            /* node read function */
int          cb_overflow();         /* node overflow function */
char         *malloc();             /* memory allocation function */



/**************************************************************************/
/*.doc *cb_alloc_node(external)*/
/*+
   This function is used to allocate memory for storing the buffer. 
-*/
/**************************************************************************/
void 
/* FCN */ **cb_alloc_node(size)
  int size;                         /*  (I) size in bytes of node  */
{
    /* ALLOCATE THE NODE */
    if ((cb = (void **)malloc(size * sizeof(void **))) == NULL) 
        return NULL;
   
    /* INITIALIZE THE POINTERS FOR THE NODE */
    rp = NULL;
    cp = NULL;
    ep = cb + (size - 1);

    /* RETURN A POINTER TO THE NODE */
    return cb;
}                                   /*  END CB_ALLOC_NODE FUNCTION  */
 
 
/**************************************************************************/
/*.doc cb_write(external)*/
/*+
   This function writes data to the buffer. The user must specify the
   information to write to the buffer. This function will find the next
   pointer in the list and write the information in the next available node.
   The user must call cb_overflow everytime they write to the buffer. If
   the function fails an error will be set.

-*/
/**************************************************************************/
int 
/* FCN */ cb_write(data)
  void *data;                       /*  (I) info to write  */
{
    /* INTIALIZE THE CURRENT POINTER TO THE BEGINNING OF THE BUFFER. 
       PLACE THE DATA IN THE CURRENT POINTER AND INCREMENT CP */
    if (cp == NULL)
        cp = cb; 
        
    *cp++ = data; 

    /* IF CURRENT POINTER IS EXCEEDS THE END POINTER, THEN 
       PLACE THE DATA AT THE BEGINNING OF THE BUFFER */
    if (cp > ep) 
        cp = cb;

    /* IF THE READ POINTER IS THE EQUAL TO THE CURRENT 
       POINTER  RETURN  BAD, OTHERWISE RETURN GOOD */
    if (rp == cp) 
        return BAD;
    
    return GOOD;
}                                   /*  END THE CB_WRITE FUNCTION  */


/**************************************************************************/
/*.doc cb_read(external)*/
/*+
   This function is used to read the next element from the buffer.  The user
   is required to specify the address where the information is stored. This
   function will return the next node in the buffer.   
-*/
/**************************************************************************/
void 
/* FCN */ *cb_read()
{
    void **sp;                      /*  (I) save pointer  */
 
    /* IF THE READ POINTER IS INTIALIZE TO NULL, THEN INTIALIZE
       THE READ POINTER TO THE BEGINNING OF THE BUFFER. */
    if (rp == NULL) 
        rp = cb;

    /* SAVE THE CURRENT READ POINTER. START READING THE LIST
       INCREMENT READ POINT TO READ THE NEXT NODE IN THE BUFFER */
    sp = rp++;

    /* WHEN READ POINTER HAS REACHED THE END OF THE
       LIST READ FROM THE BEGINNING OF THE BUFFER */
    if (rp > ep) {
        rp = cb;
    }

    /*  RETURN THE SAVE POINTER  */
    return *sp;
}                                   /*  END CB_READ FUNCTION  */


/**************************************************************************/
/*.doc cb_overflow */
/*+

   The user is responsible for calling this function everytime they write to
   the buffer and may also call it to check and see if the element they want
   to read exists in the list. If the buffer has exceed it's memory
   limitation, the user will be flagged when they call cb_overflow. It will
   return BAD if the buffer will be overwitten or GOOD otherwise. 
-*/
/*************************************************************************/
int
/* FCN */ cb_overflow()     
{
    /* IF THE CURRENT POINTER AND THE READ POINTER ARE NOT 
       INTIALIZED TO NULL, AND THE READ POINTER IS INTIALIZED 
       TO THE CURRENT POINTER, RETURN BAD. OTHERWISE RETURN GOOD. */         
    if ((cp != NULL && rp != NULL && rp == cp) || (rp == NULL && cp == cb))
        return BAD;
     
    return GOOD;

}                                   /*  END THE CB_OVERFLOW FUNCTION  */
