/**************************************************************************
   Module Name:  dlink      sdk_src_dlink_dlink.c

   Description:  Doubly Linked List Processing Library

   Author:       Tools and Technologies Group
                 Autodesk, Inc.

     (C) Copyright 1990, 1991 by Autodesk, Inc.
***************************************************************************
*  Permission to use, copy, modify, and distribute this software and its  *
*  documentation for the purpose of creating applications for AutoCAD, is *
*  hereby granted in accordance with the terms of the License Agreement	  *
*  accompanying this product.						  *
*									  *
***************************************************************************
*									  *
*  AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS.	AUTODESK  *
*  SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF MERCHANTABILITY OR	  *
*  FITNESS FOR A PARTICULAR USE.  AUTODESK, INC. DOES NOT WARRANT THAT	  *
*  THE OPERATION OF THE PROGRAM WILL BE UNINTERRUPTED OR ERROR FREE.	  *
*									  *
***************************************************************************
*									  *
*			RESTRICTED RIGHTS LEGEND			  *
*									  *
*  Use, duplication, or disclosure by the U.S. Government is subject	  *
*  to restrictions set forth in FAR 52.227-19 (Commercial Computer	  *
*  Software - Restricted Rights) and DFAR 252.227-7013 (c) (1) (ii)	  *
*  (Rights in Technical Data and Computer Software), as applicable.	  *
***************************************************************************
*									  *
*   Last Modification:							  *
*						                          *
*   Notes and restrictions on use:					  *
*    This module provides some generalised utilities for creating and     *
*    managing a doubly linked list.					  *
*									  *
*    A special structure is used here which is defined by the typedef     *
*    "dlinkype".  This structure allows the forward pointers to be        *
*    generic across all platforms.  All the user must ensure is that      *
*    the first entry in the structure to be managed by this doubly        *
*    linked list utility be a forward pointer to the next node in the     *
*    list. To guarantee portability across all platforms, the user        *
*    could easily create a union of the dlinktype structure and the       *
*    structure used in the calling program.                               *
*                                   					  *
*    A different version of this library will be provided for each of     *
*    the possible memory models. DOS can have two or four byte            *
*    pointers. The DLINK can handle both cases, but different             *
*    compilations will affect the resulant library. The user is           *
*    responsible for determining whether or not the ``far'' pointers      *
*    are required.                                                        *
*                             						  *
***************************************************************************/
/**************************************************************************/
/*  STANDARD C INCLUDE FILES  */
/**************************************************************************/
#include  <stdio.h>
#include  "sdk.h"


/**************************************************************************/
/*  GLOBAL DEFINES  */
/**************************************************************************/
#define DLINK_STORE_END          'E'         /*  store type end  */
#define DLINK_STORE_HEAD         'H'         /*  store type head  */
#define DLINK_STORE_AFTER        'A'         /*  store type after */
#define DLINK_STORE_BEFORE       'B'         /*  store type before */


/**************************************************************************/
/*  GLOBAL TYPEDEFS  */
/**************************************************************************/
typedef struct dlink {                       /*  LINKED LIST STRUCTURE  */
    struct dlink          *prev;             /*  previous pointer */ 
    struct dlink          *next;             /*  forward pointer  */
    char                  *data;             /*  data for linked list  */
                     } dlinktype;


/**************************************************************************/
/*  GLOBAL FUNCTIONS  */
/**************************************************************************/
char          *dlink_alloc_node();        /*  node allocation function */
int            dlink_delete_node();       /*  node deletion function  */
int            dlink_free_node();         /*  node free function  */
int            dlink_store_node();        /*  node store function  */
int            dlink_check_node();        /*  node check function */
char           *malloc();                 /*  memory alloc function  */
void           free();                    /* memory free function */ 


/**************************************************************************/
/*.doc dlink_alloc_node(external)*/
/*+
   This function is used to allocate a node for storing in a linked list.
   The node is completely zeroed, except for the forward pointer and the
   previous pointer which are made NULL. 
   
   Although on some systems this maybe zero, there is no guarantee that a 
   NULL pointer is zero, therefore, this is handled explicitly.  The 
   functions return a char pointer which is the buffer that was allocated.  
   The user is responsible for casting the pointer to the correct type.  If 
   NULL is returned, then the memory for the node could not be allocated.  
   The size argument is the size of the structure to be allocated.  This 
   value is in bytes and must include the size of the forward pointer.
-*/
/**************************************************************************/
char 
/* FCN */ *dlink_alloc_node(size)
  int size;                        /*  (I) size in bytes of node  */
{
    dlinktype *node;               /*  pointer to linked list structure  */

    /* ALLOCATE THE NODE */
    if ((node = (dlinktype *)malloc(size)) == NULL) 
        return NULL;
   
    /* INITIALIZE THE FORWARD POINTER FOR THE NODE */
    node->next = NULL;

    /* INITIALIZE THE PREVIOUS POINTER FOR THE NODE */
    node->prev = NULL;

    /* RETURN A POINTER TO THE NODE */
    return (char *)node;

/* END DLINK_ALLOC_NODE FUNCTION */
}


/**************************************************************************/
/*.doc dlink_delete_node(external)*/
/*+
   This function is used to delete a node from the linked list.  The linked
   list is updated to reflect the change, however, the user is responsible
   for calling dlink_free_node to ensure that the memory used by the node is
   freed correctly. The argument to the function is a pointer to the node to 
   be deleted from the list

   The function returns BAD if it couldn't locate the node in the specified
   linked list, and returns GOOD otherwise.
-*/
/**************************************************************************/
/* FCN */ dlink_delete_node(node, head)
  dlinktype *node;                  /*  (I) node to delete  */
  dlinktype **head;                 /*  (I) head of linked list  */
{
    dlinktype *n;                     /*  local node pointer  */

    /* DELETE THE SPECIFIED NODE */
    /* Check if a link list exists. If the node to be deleted 
       is the head, then make the next structure the head and 
       take the next node's pointer and direct it to the next node  
       in the list. Then take the prev address and make it NULL */
    if (*head != NULL) {               
        if (*head == node) {             
            *head = node->next;         
            node->next->prev = NULL;     
        }                                          

        /* If the node to be deleted is not the head, proceed through the 
           list until the node to be deleted is found. If the node pointer 
           reaches the end of the list(n->next ==NULL) without finding 
           the node to delete, then exit the loop with a BAD status. */
        else {
            for (n=*head; n->next!=node && n->next!=NULL; n=n->next);
                if (n->next == NULL) 
                    return BAD;

        /* If the node to be deleted is found, reference the node to be 
           deleted and get the next address of the previous structure and 
           put the contents in the next node in the list. If the next node 
           in the list is not the last link in the list, then reference 
           delete node and get the previous address of the next structure 
           and put the contents in the previous node in the list. For 
           example, if there were three links in a chain, link 1, link 2 
           and link 3 and I wanted to remove link 2, but I needed to  pass 
           the information link 3 has over to link 1. Therefore I would have 
           link 3 read my information about link 1 and copy it. */   

            node->prev->next = node->next;
            
            if (node->next != NULL) 
                node->next->prev = node->prev;
            }     
        } else 
            return BAD;

    /* RETURN GOOD STATUS TO CALLER  */
     return GOOD;

/*  END DLINK_DELETE_NODE FUNCTION  */
}


/**************************************************************************/
/*.doc dlink_free_node(external)*/
/*+
   This function is used to free a node which has been deleted from the
   linked list.  The user is responsible for deleting the node and then
   calling this function to delete the node.
-*/
/**************************************************************************/
/* FCN */ dlink_free_node(node)
  dlinktype *node;                  /*  (I) node to be freed  */
{
    /* FREE THE MEMORY ALLOCATED TO THE NODE */
    free((char *)node);

    /* RETURN GOOD STATUS TO CALLER */
    return GOOD;

/* END DLINK_FREE_NODE FUNCTION */
}

/**************************************************************************/
/*.doc dlink_check_node(internal)*/
/*+
   This function checks to see if the node address given in the 
   DLINK_STORE_BEFORE and the DLINK_STORE_AFTER functions exists in the 
   link list. dlink_check_node will start at the head of the list and 
   proceed through the list until it finds the address the user has given
   to the variable node.If the node address exists in the list the function
   will exit it's loop and return GOOD. If the node address does not exist 
   in the list then the function will return BAD.

-*/
/***************************************************************************/
/* FCN */ dlink_check_node(head, node)
  dlinktype **head;                        /*  (I) head of linked list */
  dlinktype *node;                         /*  (I) node to insert */ 
{
    dlinktype *n;                            /* local node pointer */

    /* Start at the head of the list. Proceed through the 
       link list until the address to node is found. When the 
       node pointer is equal to node address given, then return 
       a GOOD status. If the node pointer does not find the 
       node address, return a BAD status. */

   for (n=*head; n!=NULL; n=n->next) {
      if (n == node) return GOOD;
   }

   return BAD;

/* END DLINK_CHECK_NODE FUNCTION */
} 


/**************************************************************************/
/*.doc dlink_store_node(external)*/
/*+
   This function is used to store a node into the linked list.  There are
   four ways a node can be stored in the linked list.  First, it can be 
   stored at
   the HEAD of the list as would be the case in a last in first out (LIFO)
   list, which also could be a stack.  Next, it can be stored at the END
   of the list as is the case in first in first out (FIFO) lists, which
   would be the queue case. Thirdly it can be stored BEFORE a specified node 
   in the DLL if the node address is specified in place of the variable  
   nodep. Lastly, it can be stored AFTER a specified node if the node 
   address is specified in place of the variable nodep.

   The function returns BAD if an invalid type was specified, and returns
   GOOD otherwise.
-*/
/**************************************************************************/
/* FCN */ dlink_store_node(node, head, nodep, type)
  dlinktype *node;                  /*  (I) node to store  */
  dlinktype **head;                 /*  (I) head of linked list */
  dlinktype *nodep;                 /*  (I) node  pointer */
  char      type;                   /*  (I) type of store  */
{
    dlinktype *n;                     /*  local node pointer  */


    /*  STORE NODE IN LIST BASED ON THE SPECIFIED TYPE  */

    /* Check if a link list exists. If it does not exist, exit with a BAD 
       status. If a list exists, start at the head, execute the for loop 
       until n->next is equal to NULL. It has now found the end of the list.  
       If the node address is at the end of the linked list, then take 
       previous node address and assign it to the new node. Make the last 
       node address a null pointer since the node will be at the end of the
       list. If the node address is the head, store the node as the head and 
       take the next node address and assign it to head. Make the previous 
       address a null pointer. */

    switch (type) {
    case DLINK_STORE_END:
        if (*head != NULL) { 
            for (n=*head; n->next!=NULL; n=n->next);
                n->next = node;
         
            node->prev = n;
        } else {                              
           *head = node;                  
           node->prev = NULL;             
        }                                     
         
           node->next = NULL;               
           break;                              

       /* Check if a link list exists. Make the previous node address of 
          head point to node. Take the next address of node and pass 
          it to the head. Make node's previous address a null pointer. This 
          makes node the new head of the list. */                                           
    case DLINK_STORE_HEAD:
        if (*head != NULL) 
            (*head)->prev = node;
        
        node->next = *head;
        node->prev = NULL;
        *head = node;
        break;


/* If the address passed to this function is at the head, make 
   the previous node address of head point to node. Take the next 
   pointer of node and pass it to the head. Make node's previous
   address a null pointer. This makes node the new head of the list.

   Otherwise, invoke the check_node function to see if the node 
   address given exists in the list, if it does return GOOD, otherwise 
   BAD. After the function has found the node address, reference nodep 
   get the the next address of the previous structure and put the  
   contents in node. Then reference node and take the previous address 
   of node and point it to the previous structure, who's address can be 
   found in nodep. 
   
   Next reference node and point the next address of node to nodep. Then
   reference nodep and take the previous pointer of nodep and point it 
   to node. This inserts node before the nodep address that was given. */


    case DLINK_STORE_BEFORE:
        if (*head == nodep) { 
            (*head)->prev = node;
            node->next = *head;
            node->prev = NULL;
            *head = node;
        } else {
            if (dlink_check_node(head,nodep) == GOOD) {
                nodep->prev->next = node;            
                node->prev = nodep->prev;
                node->next = nodep;
                nodep->prev = node; 
            } else 
                return BAD;
       }
 
       break; 

/* If the address passed to this function is at the end 
   of the list, make the next node address of nodep point to 
   node. Take node's previous address and point it to nodep.
   Next make node's next pointer point to NULL. This makes
   node the new end of the list. 

   Otherwise, invoke the check_node function to see if the node

   address given exists in the list, if it does return GOOD, otherwise
   BAD. After the function has found the node address, reference nodep
   and get the previous address of the next structure and put the 
   contents in node. Then reference node and take the next address of
   node and point it to the next structure, whose address can be found
   in nodep. Then reference nodep and take the next pointer of nodep and
   point it to node.  This inserts node after the nodep address that was
   passed. */
  
    case DLINK_STORE_AFTER:
        if (dlink_check_node(head, nodep) == GOOD) {
            if (nodep->next == NULL) {
                nodep->next = node;
                node->prev = nodep;
                node->next = NULL;
            } else { 
                nodep->next->prev = node;
                node->next = nodep->next;
                node->prev = nodep;
                nodep->next = node;
            } 
        } else 
                return BAD;
        break;  
       
    default:
        return BAD;
    }

    return GOOD;                /* Return Success to Caller */

/*  END DLINK_STORE_NODE FUNCTION  */
}
