/***************************************************************************
   Module Name:  glads           sdk_src_gl_glads.c

   Description:  Geometry Library (ADS functions are called)

   Author     :  Tools and Technologies Group
                 Autodesk, Inc.
                 Software Developers' Kit

    (C) Copyright 1990, 1991 by Autodesk, Inc.
***************************************************************************
*  Permission to use, copy, modify, and distribute this software and its  *
*  documentation for the purpose of creating applications for AutoCAD, is *
*  hereby granted in accordance with the terms of the License Agreement   *
*  accompanying this product.                                             *
*                                                                         *
***************************************************************************
*                                                                         *
*  AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS. AUTODESK   *
*  SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF MERCHANTABILITY OR      *
*  FITNESS FOR A PARTICULAR USE.  AUTODESK, INC. DOES NOT WARRANT THAT    *
*  THE OPERATION OF THE PROGRAM WILL BE UNINTERRUPTED OR ERROR FREE.      *
*                                                                         *
***************************************************************************
*                                                                         *
*                       RESTRICTED RIGHTS LEGEND                          *
*                                                                         *
*  Use, duplication, or disclosure by the U.S. Government is subject      *
*  to restrictions set forth in FAR 52.227-19 (Commercial Computer        *
*  Software - Restricted Rights) and DFAR 252.227-7013 (c) (1) (ii)       *
*  (Rights in Technical Data and Computer Software), as applicable.       *
***************************************************************************

   Modification History:
 

   Notes and restrictions on use:
       This module provides a geometry library with functions for manipulating 
       and creating geometrical entities.  This library can be expanded any
       time later when there is a need for more functions. 
       
***************************************************************************/
/**************************************************************************/
/*  STANDARD C INCLUDE FILES  */
/**************************************************************************/

#include  <stdio.h>
#include  <math.h> 
#include  <assert.h> 
#include  <string.h>
#include  "adslib.h"
#include  "sdk.h"
#include  "gl.h"

 
/**************************************************************************/
/*  GLOBAL DEFINES  */
/**************************************************************************/
#define X        0
#define Y        1
#define Z        2     
#define LLL_STORE_END   'E'
#define LLL_STORE_HEAD  'H'
#define ANGTOL   1e-6
#define TOL      1e-10
#define GL_STRING  0
#define GL_LAYER   8
#define GL_COORD   10
#define GL_THICK   39
#define GL_COLOR   62
#define GL_NORMAL  210
 
 
/**************************************************************************/
/*  GLOBAL TYPEDEFS  */
/**************************************************************************/

typedef char       *Layer;
 
/**************************************************************************/
/*  GLOBAL VARIABLES   */
/**************************************************************************/

static struct {
               Layer layer;
               int   color;
               ads_real thick;
              } entstate = {"0",0,0.0}; 

extern char *lll_alloc_node();
extern int lll_store_node();
extern int lll_delete_node();
extern int lll_free_node();
extern pvtype    *pvhead;
extern pltype    *plhead;


/**************************************************************************/
/*.doc gl_ang_ptcirc(pt, circle)*/
/*+
   This function is used to find the angle a point makes relative to a
   circle.
-*/
/**************************************************************************/
ads_real
/* FCN */gl_ang_ptcirc(pt, circle)
ads_point  pt;                         /* (I) point */
ctype      *circle;                    /* (I) circle data */
{

   return ads_angle(circle->pt,pt);
 
/*  END GL_ANG_PTCIRC FUNCTION  */
}


/**************************************************************************/
/*.doc gl_cr_entpt(pt)*/
/*+
   This function is used to create a point entity.
-*/
/**************************************************************************/
int
/* FCN */gl_cr_entpt(pt)
ads_point    pt;                                    /* (O) point */
{
ads_name  en_name;
struct resbuf entpt,
              layrb,
              colrb,
              thkrb,
              ptrb;
struct resbuf pd;

   entpt.restype = GL_STRING;
   entpt.resval.rstring = "POINT";
   entpt.rbnext = &layrb;
   layrb.restype = GL_LAYER;
   layrb.resval.rstring = entstate.layer;
   layrb.rbnext = &colrb;
   colrb.restype = GL_COLOR;
   colrb.resval.rint = entstate.color;
   colrb.rbnext = &thkrb;
   thkrb.restype = GL_THICK;
   thkrb.resval.rreal = entstate.thick;
   thkrb.rbnext = &ptrb;
   ptrb.restype = GL_COORD;
   ptrb.resval.rpoint[X] = pt[X];
   ptrb.resval.rpoint[Y] = pt[Y];
   ptrb.resval.rpoint[Z] = pt[Z];
   ptrb.rbnext = NULL;
   ads_getvar("PDMODE",&pd);
   pd.resval.rint = 36;
   ads_setvar("PDMODE",&pd);
   ads_entmake(&entpt);
   ads_entlast(en_name);
   ads_redraw(en_name,1);
   pd.resval.rint = 0;
   ads_setvar("PDMODE",&pd);  
   return GOOD;
 
/*  END GL_CR_ENTPT FUNCTION  */
}


/**************************************************************************/
/*.doc gl_eang_ptcirc(angle, pt, circle)*/
/*+
   This function is used to compute elevation angles of tangents from a
   point to a circle.
-*/
/**************************************************************************/
int
/* FCN */gl_eang_ptcirc(angle, pt, circle)
ads_real   angle[2];                   /* (O) elevation angles */
ads_point  pt;                         /* (I) point */
ctype      *circle;                    /* (I) circle data */
{
ads_real   ang, dist,sang;

   ang = ads_angle(pt,circle->pt);
   if ((dist=gl_dpp(pt,circle->pt))<circle->radius)
      return BAD;
   sang = asin(circle->radius/dist);
   angle[0] = gl_fixang(ang+sang);
   angle[1] = gl_fixang(ang-sang);
   return GOOD;
 
/*  END GL_EANG_PTCIRC FUNCTION  */
}


/**************************************************************************/
/*.doc gl_ec2wc(ap, bp, nvec, disp)*/
/*+
   This function transforms a point bp in the Entity Coordinate System into
   the point in the World Coordinate System.
-*/
/**************************************************************************/
int
/* FCN */gl_ec2wc(ap, bp, nvec, disp)
ads_point  ap;                     /* (O) transformed point or disp. */
ads_point  bp;                     /* (I) point or disp. to be transformed */
ads_point  nvec;                   /* (I) normal vector to ECS */
int        disp;                   /* (I) flag indicating point or disp. */
{
struct resbuf rbf, rbt;

   rbf.restype = RT3DPOINT;
   rbf.resval.rpoint[X] = nvec[X];
   rbf.resval.rpoint[Y] = nvec[Y];
   rbf.resval.rpoint[Z] = nvec[Z];
   rbt.restype = RTSHORT;
   rbt.resval.rint = 0;
   if (ads_trans(bp,&rbf,&rbt,disp,ap)==RTNORM)
      return GOOD;
   else return BAD;
 
/*  END GL_EC2WC FUNCTION  */
}


/**************************************************************************/
/*.doc gl_ec2uc(ap, bp, nvec, disp)*/
/*+
   This function transforms a point bp in the Entity Coordinate System into
   the point in the User Coordinate System.
-*/
/**************************************************************************/
int
/* FCN */gl_ec2uc(ap, bp, nvec, disp)
ads_point  ap;                     /* (O) transformed point or disp. */
ads_point  bp;                     /* (I) point or disp. to be transformed */
ads_point  nvec;                   /* (I) normal vector to ECS */
int        disp;                   /* (I) flag indicating point or disp. */
{
struct resbuf rbf, rbt;

   rbf.restype = RT3DPOINT;
   rbf.resval.rpoint[X] = nvec[X];
   rbf.resval.rpoint[Y] = nvec[Y];
   rbf.resval.rpoint[Z] = nvec[Z];
   rbt.restype = RTSHORT;
   rbt.resval.rint = 1;
   if (ads_trans(bp,&rbf,&rbt,disp,ap)==RTNORM)
      return GOOD;
   else return BAD;
 
/*  END GL_EC2UC FUNCTION  */
}


/**************************************************************************/
/*.doc gl_ec2dc(ap, bp, nvec, disp)*/
/*+
   This function transforms a point bp in the Entity Coordinate System into
   the point in the Display Coordinate System.
-*/
/**************************************************************************/
int
/* FCN */gl_ec2dc(ap, bp, nvec, disp)
ads_point  ap;                     /* (O) transformed point or disp. */
ads_point  bp;                     /* (I) point or disp. to be transformed */
ads_point  nvec;                   /* (I) normal vector to ECS */
int        disp;                   /* (I) flag indicating point or disp. */
{
struct resbuf rbf, rbt;

   rbf.restype = RT3DPOINT;
   rbf.resval.rpoint[X] = nvec[X];
   rbf.resval.rpoint[Y] = nvec[Y];
   rbf.resval.rpoint[Z] = nvec[Z];
   rbt.restype = RTSHORT;
   rbt.resval.rint = 2;
   if (ads_trans(bp,&rbf,&rbt,disp,ap)==RTNORM)
      return GOOD;
   else return BAD;
 
/*  END GL_EC2DC FUNCTION  */
}


/**************************************************************************/
/*.doc gl_getline(ename, line, nvec)*/
/*+
   This function is used to get the line data from a line entity.
-*/
/**************************************************************************/
int
/* FCN */gl_getline(ename, line, nvec)
  ads_name      ename;                         /* (I) entity name */
  ltype         *line;                         /* (O) pointer to line data */
  ads_point     nvec;                          /* (O) normal vector */
{
    ads_point     p1, p2;
    struct resbuf *eb, *ebuf, *head;

    head = ebuf = ads_entget(ename);
    if (ebuf!=NULL) {
        ebuf = ebuf->rbnext;
        if (!strcmp(ebuf->resval.rstring,"LINE")) {
            for (eb = ebuf; eb->restype != 10; eb = eb->rbnext);
            setpt(p1,eb->resval.rpoint);
            eb = eb->rbnext;
            setpt(p2,eb->resval.rpoint);
            for (eb = eb; eb->restype != 210; eb = eb->rbnext);
            setpt(nvec,eb->resval.rpoint);
         
            if (gl_line(line,p1,p2) == RETURN_NORMAL) { 
                ads_relrb(head); 
                return GOOD;
            } else {
                ads_relrb(head);
                return BAD;
            }
        } else {
            ads_relrb(head); 
            return BAD;
        }     
    } else {
        ads_relrb(head);
        return BAD;
    }
/*  END GL_GETLINE FUNCTION  */
}


/**************************************************************************/
/*.doc gl_getcircle(ename, circle, nvec)*/
/*+
   This function is used to get center and radius of a circle entity.
-*/
/**************************************************************************/
int
/* FCN */gl_getcircle(ename, circle, nvec)
  ads_name   ename;                        /* (I) entity name */
  ctype      *circle;                      /* (O) pointer to circle data */
  ads_point  nvec;                         /* (O) normal vector */ 
{
    struct resbuf *eb, *ebuf, *head;

    head = ebuf = ads_entget(ename);
   
    if (ebuf != NULL) {
        ebuf = ebuf->rbnext;
      
        if (!strcmp(ebuf->resval.rstring,"CIRCLE")) {
            for (eb=ebuf; eb->restype!=10; eb=eb->rbnext);
         
            setpt(circle->pt,eb->resval.rpoint);
         
            for (eb=eb->rbnext; eb->restype!=40; eb=eb->rbnext);
         
            circle->radius = eb->resval.rreal;
         
            for (eb=eb->rbnext; eb->restype!=210; eb=eb->rbnext);
         
            setpt(nvec,eb->resval.rpoint);
            ads_relrb(head);
            return GOOD;
        } else {
            ads_relrb(head);
            return BAD;
        } 
    } else {  
        ads_relrb(head); 
        return BAD;
    }
/*  END GL_GETCIRCLE FUNCTION  */
}


/**************************************************************************/
/*.doc gl_getpvtx(ename)*/ 
/*+
   This function is used to get data of vertices from a polyline entity.
-*/
/**************************************************************************/
int
/* FCN */gl_getpvtx(ename)
  ads_name      ename;                             /* (I) entity name */
{
    struct resbuf *head, *eb, *eb1, *eb2, *ebuf;
    pvtype        *pvnode;
    ads_point     pt, res, nvec;
    int           i, close, disp;
    ads_name      elocal;

    ebuf = head = ads_entget(ename);
   
    for (; ebuf->restype!=0; ebuf=ebuf->rbnext);
   
    if (strcmp(ebuf->resval.rstring, "POLYLINE") != 0) {
        ads_relrb(head);
        return BAD;
    }
 
    for (eb=ebuf; eb->restype!=70; eb=eb->rbnext);
    
    close = (eb->resval.rint==1) ? 1 : 0;
   
    for (eb=eb; eb->restype!=210; eb=eb->rbnext);
   
    setpt(nvec,eb->resval.rpoint);
    ads_relrb(head);
   
    ads_entnext(ename, elocal);
    ebuf = ads_entget(elocal);
    eb = ebuf->rbnext; 
    
    i = 0;
    disp = 0;
    
    while (strcmp(eb->resval.rstring,"SEQEND")) {
        for (eb=eb; eb->restype!=10; eb=eb->rbnext);
        for (eb1=eb; eb1->restype!=70; eb1=eb1->rbnext);
      
        if (eb1->resval.rint!=16) {
            if ((pvnode=(pvtype *)lll_alloc_node(sizeof(pvtype)))==NULL) {
                ads_relrb(ebuf); 
                return BAD;
            }

            i += 1;
            pvnode->next = NULL;
            setpt(res,eb->resval.rpoint);
            gl_ec2uc(res,res,nvec,disp);
            setpt(pvnode->vtx,res);
            
            for (eb2=eb; eb2->restype!=42; eb2=eb2->rbnext);
          
            pvnode->bulge = eb2->resval.rreal;
         
            if (lll_store_node(pvnode,&pvhead,LLL_STORE_END)==
                               RETURN_ABNORMAL) { 
                ads_relrb(ebuf); 
                return BAD;
            }
  
            if (i==1 && close) 
                setpt(pt,pvnode->vtx);
        }
      
        ads_entnext(elocal, elocal);
        ads_relrb(ebuf);
        ebuf = ads_entget(elocal);
        eb = ebuf->rbnext;
    }

    ads_relrb(ebuf);
   
    if (close) {
        if ((pvnode=(pvtype *)lll_alloc_node(sizeof(pvtype)))==NULL)
            return BAD;
        
        setpt(pvnode->vtx,pt);
        pvnode->next = NULL;
        pvnode->bulge = 9999.0;        /* closed loop, same as the first vertex */
      
        if (lll_store_node(pvnode,&pvhead,LLL_STORE_END)==RETURN_ABNORMAL) 
            return BAD;
    }
    
    return GOOD;

/*  END GL_GETPVTX FUNCTION  */
}
 
 
/**************************************************************************/
/*.doc gl_getrpvtx(ename)*/
/*+
   This function is used to reverse the order of the vertices of a
   polyline entity.
-*/
/**************************************************************************/
int
/* FCN */gl_getrpvtx(ename)
  ads_name      ename;                              /* (I) entity name */
{
    struct resbuf *head, *eb, *ebuf, *eb1, *eb2;
    pvtype        *pvnode;
    int           i, close;
    ads_real      blg;
    ads_point     pt;
    ads_name      nname;
     
    ebuf = head = ads_entget(ename);
    
    for (; ebuf->restype!=0; ebuf=ebuf->rbnext);
   
    if (strcmp(ebuf->resval.rstring, "POLYLINE") != 0) {
        ads_relrb(head);
        return BAD;
    }
    
    for (eb=ebuf; eb->restype!=70; eb=eb->rbnext);
   
    close = (eb->resval.rint==1) ? 1 : 0;
    ads_relrb(head);
    ads_entnext(ename,nname);
    ebuf = ads_entget(nname);
    eb = ebuf->rbnext;
    i = 0; 
    blg = 0.0;
   
    while (strcmp(eb->resval.rstring,"SEQEND")) {
        for (eb=eb; eb->restype!=10; eb=eb->rbnext);
        for (eb1=eb; eb1->restype!=42; eb1=eb1->rbnext);
        for (eb2=eb1; eb2->restype!=70; eb2=eb2->rbnext);
      
        if (eb2->resval.rint!=16) {
            i += 1;
            
            if (i==1 && close) {
                setpt(pt,eb->resval.rpoint);
            } else {
                if ((pvnode=(pvtype *)lll_alloc_node(sizeof(pvtype)))==NULL) {
                    ads_relrb(ebuf);
                    return BAD;
                }
 
                setpt(pvnode->vtx,eb->resval.rpoint);
                pvnode->bulge = blg;
                pvnode->next = NULL;
                
                if (lll_store_node(pvnode,&pvhead,LLL_STORE_HEAD)==BAD) {
                    ads_relrb(ebuf); 
                    return BAD;
                }
            }
         blg = -1.0*eb1->resval.rreal;
        }
      
        ads_entnext(nname,nname);
        ads_relrb(ebuf);
        ebuf = ads_entget(nname);
        eb = ebuf->rbnext;
    }
   
    ads_relrb(ebuf);
   
    if (close) {
        if ((pvnode=(pvtype *)lll_alloc_node(sizeof(pvtype)))==NULL)
            return BAD;
        pvnode->bulge = blg;
        setpt(pvnode->vtx,pt);
        pvnode->next = NULL;
      
        if (lll_store_node(pvnode,&pvhead,LLL_STORE_HEAD)==BAD)
            return BAD;
    }      
   
    return GOOD;
 
/*  END GL_GETRPVTX FUNCTION  */
}


/**************************************************************************/
/*.doc gl_getpline(ename)*/
/*+
   This function is used to retrieve the line segments from a polyline
   entity.
-*/
/**************************************************************************/
int
/* FCN */gl_getpline(ename)
  ads_name      ename;                               /* (I) entity name */
{
    struct resbuf *ebuf, *eb, *head;

    head = ebuf = ads_entget(ename);

    if (ebuf != NULL) {
        eb = ebuf->rbnext;
      
        if (!strcmp(eb->resval.rstring,"POLYLINE")) {
            ads_relrb(head);
         
            if (gl_getpvtx(ename)==BAD) 
                return BAD;
         
            if (gl_p2lines() == GOOD) 
                return GOOD;
         
            return BAD;
        }
      
      ads_relrb(head); 
      return BAD;
    }  
    
    ads_relrb(head); 
    return BAD;

/*  END GL_GETPLINE FUNCTION  */
}


/**************************************************************************/
/*.doc gl_getrpline(ename)*/
/*+
   This function is used to reverse the line segments from a polyline
   entity.
-*/
/**************************************************************************/
int
/* FCN */gl_getrpline(ename)
  ads_name      ename;                              /* (I) entity name */
{
    struct resbuf *ebuf, *eb, *head;

    head = ebuf = ads_entget(ename);
   
    if (ebuf != NULL) {
        eb = ebuf->rbnext;
      
        if (!strcmp(eb->resval.rstring,"POLYLINE")) {
            ads_relrb(ebuf);
         
        if (gl_getrpvtx(ename)==BAD) 
            return BAD;
         
        return GOOD;
        }
     
      ads_relrb(head); 
      return BAD;
    }
    
    ads_relrb(head); 
    return BAD;

/*  END GL_GETRPLINE FUNCTION  */
}


/**************************************************************************/
/*.doc gl_get_line(prompt, ename, line, nvec)*/
/*+
   This function is used to test the function of getting the line data.
-*/
/**************************************************************************/
int
/* FCN */gl_get_line(prompt, ename, line, nvec)
char          *prompt;                       /* (I) prompt */
ads_name      ename;                         /* (O) entity name */
ltype         *line;                         /* (O) pointer to line data */
ads_point     nvec;                          /* (O) normal vector */
{
ads_point     ipt;
int           ind;

   ind = 0;
   ipt[X] = ipt[Y] = ipt[Z] = 0.0;
   ename[0] = ename[1] = 0;
   while (ind == 0) {
      if (ads_entsel(prompt,ename,ipt)==RTCAN) return BAD;
      if (gl_getline(ename,line,nvec)==GOOD) ind = 1;
      if (ind == 0) ads_printf("Not a valid line entity.\n");
   }
   return GOOD;

/*  END GL_GET_LINE FUNCTION  */
}


/**************************************************************************/
/*.doc gl_get_circle(prompt, circle, nvec)*/
/*+
   This function is used to test the function of getting the circle data.
-*/
/**************************************************************************/
int
/* FCN */gl_get_circle(prompt, circle, nvec)
char          *prompt;                          /* (I) entity name */
ctype         *circle;                          /* (O) pointer to circle data */
ads_point     nvec;                             /* (O) normal vector */
{
ads_point     ipt;
ads_name      ename;
int           ind;

   ind = 0;
   ipt[X] = ipt[Y] = ipt[Z] = 0.0;
   ename[0] = ename[1] = 0;
   while (ind == 0) {
      if (ads_entsel(prompt,ename,ipt) == RTCAN) return BAD;
      if (gl_getcircle(ename,circle,nvec)==GOOD) ind = 1;
      if (ind == 0) ads_printf("Not a valid circle entity.\n");
   }
   return GOOD;

/*  END GL_GET_CIRCLE FUNCTION  */
}


/**************************************************************************/
/*.doc gl_get_pline(prompt)*/
/*+
   This function is used to test the function of retrieving the polyline
   entity data.
-*/
/**************************************************************************/
int
/* FCN */gl_get_pline(prompt)
char          *prompt;                            /* (I) prompt */
{
ads_point     ipt;
ads_name      ename;
int           ind;

   ind = 0;
   ipt[X] = ipt[Y] = ipt[Z] = 0.0;
   ename[0] = ename[1] = 0;
   while (ind == 0) {
      if (ads_entsel(prompt,ename,ipt) == RTCAN) return BAD;
      if (gl_getpline(ename)==GOOD) ind = 1;
      if (ind == 0) ads_printf("Not a valid polyline entity.\n");
   }
   return GOOD;

/*  END GL_GET_PLINE FUNCTION  */
}
 
 
/**************************************************************************/
/*.doc gl_set_rpline(prompt)*/
/*+
   This function is used to get input data of a polyline in order to
   reverse its representation.
-*/
/**************************************************************************/
int
/* FCN */gl_set_rpline(prompt)
char          *prompt;                             /* (I) prompt */
{
ads_point     ipt;
ads_name      ename;
int           ind;
pvtype        *pvnode;

   ind = 0;
   ipt[X] = ipt[Y] = ipt[Z] = 0.0;
   ename[0] = ename[1] = 0;
   while (ind == 0) {
      if (ads_entsel(prompt,ename,ipt) == RTCAN) return BAD;
      if (gl_getrpline(ename)==GOOD) ind = 1;
      if (ind == 0) ads_printf("Not a valid polyline entity.\n");
   }
   if (gl_rpline(ename)==BAD) return BAD;
   ads_printf("The polyline has been reversed.\n");
   while (pvhead!=NULL) {
      pvnode = pvhead;
      if (lll_delete_node(pvnode,&pvhead)==BAD) return BAD;
      lll_free_node(pvnode);
   }
   return GOOD;

/*  END GL_GET_RPLINE FUNCTION  */
}


/**************************************************************************/
/*.doc gl_intlc_spc(pts, line, circle, nvl, nvec)*/
/*+
   This function is used to calculate the intersection point(s) of a line
   and a circle on a plane in 3D space (arbitrary coordinate system). 
-*/
/**************************************************************************/
int
/* FCN */gl_intlc_spc(pts, line, circle, nvl, nvec)
ads_point     pts[2];                       /* (O) intersection point(s) */
ltype         *line;                        /* (I) pointer to line data */
ctype         *circle;                      /* (I) pointer to circle data */
ads_point     nvl;                          /* (I) normal vector to the plane 
                                                   where the line lies */
ads_point     nvec;                         /* (I) normal vector to the plane 
                                                   where the circle lies */
{
ads_point     pt, res1, res2;
int           i, j, n, disp;

   if (!gl_vec_eqv(nvl,nvec)) return -1;
   disp = 0;
   gl_wc2uc(res1,line->pt1,disp);
   gl_wc2uc(res2,line->pt2,disp);
   gl_line_coef(line,res1,res2);
   gl_ec2uc(res1,circle->pt,nvec,disp);
   setpt(circle->pt,res1);
   n = gl_intlc_pln(pts,line,circle); 
   for (i=0; i<n; i++) {
      for (j=0; j<3; j++) pt[j] = pts[i][j];
      gl_uc2wc(pt,pt,disp);
      for (j=0; j<3; j++) pts[i][j] = pt[j];
   }
   return n;
 
/*  END GL_INTLC_SPC FUNCTION  */
}


/**************************************************************************/
/*.doc gl_intlc_proj(pts, line, circle, nvec)*/
/*+
   This function is used to calculate the sliding apparent intersection
   point(s) of a line and a circle on a viewing plane when the circle and
   the line do not lie on the same plane.
-*/
/**************************************************************************/
int
/* FCN */gl_intlc_proj(pts, line, circle, nvec)
ads_point     pts[2];                          /* (O) intersection point(s) */
ltype         *line;                           /* (I) pointer to the line data */
ctype         *circle;                         /* (I) pointer to the circle data */
ads_point     nvec;                            /* (I) normal vector to the plane
                                                      where the circle lies */
{
cotype        conic;
ltype         line1;
ads_point     res, pt;
ads_real      sum;
register int  i;
int           n, disp;
txform        invm, m;

   disp = 0;
   gl_line_coef(line,line->pt1,line->pt2);
   if (gl_ac2wc(m,nvec)==BAD) return BAD;
   gl_ec2wc(res,circle->pt,nvec,disp);
   for (i=0; i<3; i++) m[i][3] = res[i];
   if (gl_mat_inv(invm,m)==BAD) return BAD;
   if (fabs(invm[2][2])<=TOL) {
      line1.g = invm[2][0];
      line1.h = invm[2][1];
      line1.i = invm[2][3];
      if (gl_intll_coef(pt,line,&line1)==BAD) return BAD;
      sum = 0.0;
      for (i=0; i<3; i++) sum = sum + (pt[i]-res[i])*(pt[i]-res[i]);
      if (sum<=(circle->radius*circle->radius)) {
         for (i=0; i<3; i++) pts[0][i] = pt[i];
         n = 1;
      }
      else n = 0;
   }
   else {
      gl_cir2eli(&conic,invm,circle);
      n = gl_int_lco(pts,line,&conic); 
   }
   return n;
 
/*  END GL_INTLC_PROJ FUNCTION  */
}
 

/**************************************************************************/
/*.doc gl_intcc_spc(pts, circ1, circ2, nv1, nv2)*/
/*+
   This function is used to calculate the intersection point(s) of two
   circles.
-*/
/**************************************************************************/
int
/* FCN */gl_intcc_spc(pts, circ1, circ2, nv1, nv2)
ads_point  pts[2];                     /* (O) intersection points */
ctype      *circ1, *circ2;             /* (I) circles data */
ads_point  nv1, nv2;
{
ads_point  res1, res2, pt;
int        i, j, n, disp; 
 
   if (!gl_vec_eqv(nv1,nv2)) return -1;
   disp = 0;
   gl_ec2uc(res1,circ1->pt,nv1,disp);
   gl_ec2uc(res2,circ2->pt,nv1,disp);
   setpt(circ1->pt,res1);
   setpt(circ2->pt,res2);
   n = gl_intcc_pln(pts,circ1,circ2); 
   for (i=0; i<n; i++) {
      for (j=0; j<3; j++) pt[j] = pts[i][j];
      gl_uc2wc(pt,pt,disp);
      for (j=0; j<3; j++) pts[i][j] = pt[j];
   }
   return n;
 
/*  END GL_INTCC_SPC FUNCTION  */
}


/**************************************************************************/
/*.doc gl_int_ll(pt, line1, line2)*/
/*+
   This function is used to calculate the intersection point of two
   lines in 3D space.
-*/
/**************************************************************************/
int
/* FCN */gl_int_ll(pt, line1, line2)
ads_point  pt;                         /* (O) a point */
ltype      *line1, *line2;             /* (I) line data */
{
ads_point  from1, to1, from2, to2;
int        teston;

   setpt(from1,line1->pt1);
   setpt(to1,line1->pt2);
   setpt(from2,line2->pt1);
   setpt(to2,line2->pt2);
   teston = 0;
   if (ads_inters(from1,to1,from2,to2,teston,pt)==RTNORM)
      return GOOD;
   else return BAD;
 
/*  END GL_INT_LL FUNCTION  */
}


/**************************************************************************/
/*.doc gl_int_pline(line1)*/
/*+
   This function is used to calculate the intersection point(s) of a line and
   a polyline.  The relevant data of the polyline is passed to this function 
   through a global head "plhead" of the linked list.
-*/
/**************************************************************************/
int
/* FCN */gl_int_pline(line1)
ltype      *line1;                      /* (I) line information */
{
ads_point  pt;
ads_point  pts[2];
int        i, j, n, m, disp;
pvtype     *pvnode;
pltype     *plnode;

   disp = 0;
   gl_wc2uc(line1->pt1,line1->pt1,disp);
   gl_wc2uc(line1->pt2,line1->pt2,disp);
   for (plnode=plhead; plnode!=NULL; plnode=plnode->next) {
      if (!plnode->bf) {
         if (gl_int_ll(pt,line1,&plnode->line)==BAD) return BAD;
         if (gl_ptseg(&plnode->line,pt)==TRUE) {
            gl_uc2wc(pt,pt,disp);
            gl_cr_entpt(pt);
         }
      }
      else {
         gl_line_coef(line1,line1->pt1,line1->pt2);
         n = gl_intlc_pln(pts,line1,&plnode->circle);
         m = gl_pline_ptarc(plnode,pts,n);
         for (i=0; i<m; i++) {
            for (j=0; j<3; j++) pt[j] = pts[i][j];
            gl_uc2wc(pt,pt,disp);
            gl_cr_entpt(pt);
         }
      }
   }
   while (pvhead!=NULL) {
      pvnode = pvhead;
      if (lll_delete_node(pvnode,&pvhead)==BAD) return BAD;
      lll_free_node(pvnode);
   }
   while (plhead!=NULL) {
      plnode = plhead;
      if (lll_delete_node(plnode,&plhead)==BAD) return BAD;
      lll_free_node(plnode);
   }
   return GOOD;
 
/*  END GL_INT_PLINE FUNCTION  */
}


/**************************************************************************/
/*.doc gl_pline_ptarc(plnode, pts, n)*/
/*+
   This function is used to check whether or not the intersection point(s)
   of a line and a circle which contains an arc of a polyline to be on the
   arc segment of the polyline.
-*/
/**************************************************************************/
int
/* FCN */gl_pline_ptarc(plnode, pts, n)
pltype     *plnode;                     /* (I) polyline node information */
ads_point  pts[2];                      /* (I/O) points on an arc? */
int        n;                           /* (I/O) number of points on an arc */
{
ads_point  pt;
ads_real   starta, enda, tempa;
ltype      *ln;
int        i, j, m, on, nm;

   ln = &plnode->line;
   starta = gl_fixang(gl_ang_ptcirc(ln->pt1,&plnode->circle));
   enda = gl_fixang(gl_ang_ptcirc(ln->pt2,&plnode->circle));
   if (plnode->bf==-1) {
      tempa = starta;
      starta = enda;
      enda = tempa;
   }
   m = n;
   for (i=0; i<n; i++) {
      for (j=0; j<3; j++) pt[j] = pts[i][j];
      on = gl_ptarc(starta,enda,pt,&plnode->circle);
      if (on) {
         nm = i-n+m;
         for (j=0; j<3; j++) pts[nm][j] = pt[j]; 
      }
      else m -= 1;
   }
   return m;
 
/*  END GL_PLINE_PTARC FUNCTION  */
}


/**************************************************************************/
/*.doc gl_ptarc(starta, enda, pt, circle)*/
/*+
   This function is used to test point(s) to be on an arc segment or not.
-*/
/**************************************************************************/
int
/* FCN */gl_ptarc(starta, enda, pt, circle)
ads_real   starta, enda;                /* (I) start angle and end angle */
ads_point  pt;                          /* (I) point on an arc? */
ctype      *circle;                     /* (I) circle data */
{
ads_real   ang;
int        on;

  ang = gl_fixang(gl_ang_ptcirc(pt,circle));
  on = (starta<enda) ? (starta-ANGTOL<=ang && ang<=enda+ANGTOL) :
                       (ang>=starta-ANGTOL || ang<=enda+ANGTOL);
  if (on) return TRUE;
  else return FALSE;
 
/*  END GL_PTARC FUNCTION  */
}


/**************************************************************************/
/*.doc gl_rpline(ename)*/
/*+
   This function modifies the data of a polyline entity in order to reverse
   the order of its representation.
-*/
/**************************************************************************/
int
/* FCN */gl_rpline(ename)
ads_name      ename;                               /* (I) entity name */
{
pvtype        *pvnode;
struct resbuf *ebuf, *eb, *eb1, *eb2;
ads_name      nname;
int           ind;

   ads_entnext(ename,nname);
   ebuf = ads_entget(nname);
   pvnode = pvhead;
   for (pvnode=pvnode; pvnode!=NULL; pvnode=pvnode->next) {
      ind = 0;
      while (ind==0) {
         for (eb2=ebuf; eb2->restype!=70; eb2=eb2->rbnext);
         if (eb2->resval.rint!=16) {
            for (eb=ebuf; eb->restype!=10; eb=eb->rbnext);
            for (eb1=eb; eb1->restype!=42; eb1=eb1->rbnext);
            setpt(eb->resval.rpoint,pvnode->vtx);
            eb1->resval.rreal = pvnode->bulge;
            if (ads_entmod(ebuf)!=RTNORM) {
               ads_printf("\nBAD polyline reverse modification.");
               ads_relrb(ebuf);
               return BAD;
            }
            ind = 1;
         }
         ads_entnext(nname,nname);
         ads_relrb(ebuf);
         ebuf = ads_entget(nname);
      }   
   }
   ads_relrb(ebuf);
   if (ads_entupd(ename)!=RTNORM) {
      ads_printf("\nBad redraw of the reverse polyline.");
      return BAD;
   }
   return GOOD;
 
/*  END GL_RPLINE FUNCTION  */
}


/**************************************************************************/
/*.doc gl_transform(m, rbf, rbt)*/
/*+
   This function creates the transformation matrix from one coordinate
   system to another coordinate system.
-*/
/**************************************************************************/
int
/* FCN */gl_transform(m, rbf, rbt)
txform     m;                          /* (O) transformation matrix */
struct resbuf *rbf, *rbt;              /* (I) result buffers */
{
int        i, disp;
ads_point  ap, bp, xdir, ydir, zdir;

   gl_mat_ident(m);
   disp = 0;
   bp[0] = 0.0; bp[1] = 0.0; bp[2] = 0.0;
   if (ads_trans(bp,rbf,rbt,disp,ap)!=RTNORM) return BAD;
   disp = 1;
   bp[0] = 1.0; bp[1] = 0.0; bp[2] = 0.0;
   if (ads_trans(bp,rbf,rbt,disp,xdir)!=RTNORM) return BAD;
   bp[0] = 0.0; bp[1] = 1.0; bp[2] = 0.0;
   if (ads_trans(bp,rbf,rbt,disp,ydir)!=RTNORM) return BAD;
   bp[0] = 0.0; bp[1] = 0.0; bp[2] = 1.0;
   if (ads_trans(bp,rbf,rbt,disp,zdir)!=RTNORM) return BAD;
   for (i=0; i<3; i++) {
      m[i][0] = xdir[i];
      m[i][1] = ydir[i];
      m[i][2] = zdir[i];
      m[i][3] = ap[i];
   }
   return GOOD;

/*  END GL_TRANSFORM FUNCTION  */
}


/**************************************************************************/
/*.doc gl_txf_dc2pc(m)*/
/*+
   This function creates the transformation matrix from Model Space Display
   Coordinate System to Paper Space Display Coordinate System.
-*/
/**************************************************************************/
int
/* FCN */gl_txf_dc2pc(m)
txform     m;                          /* (O) transformation matrix */
{
struct resbuf rbf, rbt;

   rbf.restype = RTSHORT;
   rbf.resval.rint = 2;
   rbt.restype = RTSHORT;
   rbt.resval.rint = 3;
   if (gl_transform(m,&rbf,&rbt)==BAD) return BAD;
   return GOOD;

/*  END GL_TXF_DC2PC FUNCTION  */
}


/**************************************************************************/
/*.doc gl_txf_dc2wc(m)*/
/*+
   This function creates the transformation matrix from Dispaly Coordinate
   System to World Coordinate System.
-*/
/**************************************************************************/
int
/* FCN */gl_txf_dc2wc(m)
txform     m;                          /* (O) transformation matrix */
{
struct resbuf rbf, rbt;

   rbf.restype = RTSHORT;
   rbf.resval.rint = 2;
   rbt.restype = RTSHORT;
   rbt.resval.rint = 0;
   if (gl_transform(m,&rbf,&rbt)==BAD) return BAD;
   return GOOD;

/*  END GL_TXF_DC2WC FUNCTION  */
}


/**************************************************************************/
/*.doc gl_txf_ec2wc(m, ename)*/
/*+
   This function creates the transformation matrix from Entity Coordinate
   System to World Coordinate System.
-*/
/**************************************************************************/
int
/* FCN */gl_txf_ec2wc(m, ename)
  txform     m;                          /* (O) transformation matrix */
  ads_name   ename;                      /* (I) entity name */
{
    int        nv;
    ads_point  ndir;
    struct resbuf *ebuf, *head;

    head = ebuf = ads_entget(ename);
   
    if (ebuf==NULL) {
        ads_relrb(head);
        return BAD;
    }

    for (nv=0; ebuf!=NULL && nv<1;) {
        switch(ebuf->restype)
        {
        case GL_NORMAL:
            setpt(ndir,ebuf->resval.rpoint);
            nv++;
            break;
        }
      ebuf = ebuf->rbnext;
    }
   
    if (gl_wc2ac(m,ndir)==BAD) {
        ads_relrb(head);
        return BAD;
    }
  
    ads_relrb(head); 
    return GOOD;

/*  END GL_TXF_EC2WC FUNCTION  */
}


/**************************************************************************/
/*.doc gl_txf_pc2dc(m)*/
/*+
   This function creates the transformation matrix from Paper Space Dispaly
   Coordinate System to Model Space Display Coordinate System.
-*/
/**************************************************************************/
int
/* FCN */gl_txf_pc2dc(m)
txform     m;                          /* (O) transformation matrix */
{
struct resbuf rbf, rbt;

   rbf.restype = RTSHORT;
   rbf.resval.rint = 3;
   rbt.restype = RTSHORT;
   rbt.resval.rint = 2;
   if (gl_transform(m,&rbf,&rbt)==BAD) return BAD;
   return GOOD;

/*  END GL_TXF_PC2DC FUNCTION  */
}


/**************************************************************************/
/*.doc gl_txf_wc2dc(m)*/
/*+
   This function creates the transformation matrix from World Coordinate
   System to Display Coordinate System.
-*/
/**************************************************************************/
int
/* FCN */gl_txf_wc2dc(m)
txform     m;                          /* (O) transformation matrix */
{
struct resbuf rbf, rbt;

   rbf.restype = RTSHORT;
   rbf.resval.rint = 0;
   rbt.restype = RTSHORT;
   rbt.resval.rint = 2;
   if (gl_transform(m,&rbf,&rbt)==BAD) return BAD;
   return GOOD;

/*  END GL_TXF_WC2DC FUNCTION  */
}


/**************************************************************************/
/*.doc gl_txf_wc2ec(m, ename)*/
/*+
   This function creates the transformation matrix from World Coordinate
   System to Entity Coordinate System.
-*/
/**************************************************************************/
int
/* FCN */gl_txf_wc2ec(m, ename)
txform     m;                          /* (O) transformation matrix */
ads_name   ename;                      /* (I) entity name */
{
txform     m1;

   if (gl_txf_ec2wc(m1,ename)==BAD) return BAD;
   if (gl_mat_inv(m,m1)==BAD) return BAD;
   return GOOD;

/*  END GL_TXF_WC2EC FUNCTION  */
}


/**************************************************************************/
/*.doc gl_uc2wc(ap, bp, disp)*/
/*+
   This function transforms a point bp in the User Coordinate System into
   the point in the World Coordinate System.
-*/
/**************************************************************************/
int
/* FCN */gl_uc2wc(ap, bp, disp)
ads_point  ap;                     /* (O) transformed point or disp. */
ads_point  bp;                     /* (I) point or disp. to be transformed */
int        disp;                   /* (I) flag indicating point or disp. */
{
struct resbuf rbf, rbt;

   rbf.restype = RTSHORT;
   rbf.resval.rint = 1;
   rbt.restype = RTSHORT;
   rbt.resval.rint = 0;
   if (ads_trans(bp,&rbf,&rbt,disp,ap)==RTNORM)
      return GOOD;
   else return BAD;
 
/*  END GL_UC2WC FUNCTION  */
}


/**************************************************************************/
/*.doc gl_uc2ec(ap, bp, nvec, disp)*/
/*+
   This function transforms a point bp in the User Coordinate System into
   the point in the Entity Coordinate System.
-*/
/**************************************************************************/
int
/* FCN */gl_uc2ec(ap, bp, nvec, disp)
ads_point  ap;                     /* (O) transformed point or disp. */
ads_point  bp;                     /* (I) point or disp. to be transformed */
ads_point  nvec;                   /* (I) normal vector to ECS */
int        disp;                   /* (I) flag indicating point or disp. */
{
struct resbuf rbf, rbt;

   rbf.restype = RTSHORT;
   rbf.resval.rint = 1;
   rbt.restype = RT3DPOINT;
   rbt.resval.rpoint[X] = nvec[X];
   rbt.resval.rpoint[Y] = nvec[Y];
   rbt.resval.rpoint[Z] = nvec[Z];
   if (ads_trans(bp,&rbf,&rbt,disp,ap)==RTNORM)
      return GOOD;
   else return BAD;
 
/*  END GL_UC2EC FUNCTION  */
}


/**************************************************************************/
/*.doc gl_uc2dc(ap, bp, disp)*/
/*+
   This function transforms a point bp in the User Coordinate System into
   the point in the Display Coordinate System.
-*/
/**************************************************************************/
int
/* FCN */gl_uc2dc(ap, bp, disp)
ads_point  ap;                     /* (O) transformed point or disp. */
ads_point  bp;                     /* (I) point or disp. to be transformed */
int        disp;                   /* (I) flag indicating point or disp. */
{
struct resbuf rbf, rbt;

   rbf.restype = RTSHORT;
   rbf.resval.rint = 1;
   rbt.restype = RTSHORT;
   rbt.resval.rint = 2;
   if (ads_trans(bp,&rbf,&rbt,disp,ap)==RTNORM)
      return GOOD;
   else return BAD;
 
/*  END GL_UC2DC FUNCTION  */
}


/**************************************************************************/
/*.doc gl_wc2uc(ap, bp, disp)*/
/*+
   This function transforms a point bp in the World Coordinate System into
   the point in the User Coordinate System.
-*/
/**************************************************************************/
int
/* FCN */gl_wc2uc(ap, bp, disp)
ads_point  ap;                     /* (O) transformed point */
ads_point  bp;                     /* (I) point to be transformed */
int        disp;                   /* (I) flag indicating point or vector */
{
struct resbuf rbf, rbt;

   rbf.restype = RTSHORT;
   rbf.resval.rint = 0;
   rbt.restype = RTSHORT;
   rbt.resval.rint = 1;
   if (ads_trans(bp,&rbf,&rbt,disp,ap)==RTNORM)
      return GOOD;
   else return BAD;
  
/*  END GL_WC2UC FUNCTION  */
}


/**************************************************************************/
/*.doc gl_wc2ec(ap, bp, nvec, disp)*/
/*+
   This function transforms a point bp in the World Coordinate System into
   the point in the Entity Coordinate System.
-*/
/**************************************************************************/
int
/* FCN */gl_wc2ec(ap, bp, nvec, disp)
ads_point  ap;                     /* (O) transformed point or disp. */
ads_point  bp;                     /* (I) point or disp. to be transformed */
ads_point  nvec;                   /* (I) normal vector to ECS */
int        disp;                   /* (I) flag indicating point or disp. */
{
struct resbuf rbf, rbt;

   rbf.restype = RTSHORT;
   rbf.resval.rint = 0;
   rbt.restype = RT3DPOINT;
   rbt.resval.rpoint[X] = nvec[X];
   rbt.resval.rpoint[Y] = nvec[Y];
   rbt.resval.rpoint[Z] = nvec[Z];
   if (ads_trans(bp,&rbf,&rbt,disp,ap)==RTNORM)
      return GOOD;
   else return BAD;
  
/*  END GL_WC2EC FUNCTION  */
}


/**************************************************************************/
/*.doc gl_wc2dc(ap, bp, disp)*/
/*+
   This function transforms a point bp in the World Coordinate System into
   the point in the Display Coordinate System.
-*/
/**************************************************************************/
int
/* FCN */gl_wc2dc(ap, bp, disp)
ads_point  ap;                     /* (O) transformed point or disp. */
ads_point  bp;                     /* (I) point or disp. to be transformed */
int        disp;                   /* (I) flag indicating point or disp. */
{
struct resbuf rbf, rbt;

   rbf.restype = RTSHORT;
   rbf.resval.rint = 0;
   rbt.restype = RTSHORT;
   rbt.resval.rint = 2;
   if (ads_trans(bp,&rbf,&rbt,disp,ap)==RTNORM)
      return GOOD;
   else return BAD;
 
/*  END GL_WC2DC FUNCTION  */
}


