/***************************************************************************
   Module Name:  glnads           sdk_src_gl_glnads.c

   Description:  Geometry Library (Independent of ADS functions)

   Author     :  Tools and Technologies Group
                 Autodesk, Inc.
                 Software Developers' Kit

    (C) Copyright 1990, 1991 by Autodesk, Inc.
***************************************************************************
*  Permission to use, copy, modify, and distribute this software and its  *
*  documentation for the purpose of creating applications for AutoCAD, is *
*  hereby granted in accordance with the terms of the License Agreement   *
*  accompanying this product.                                             *
*                                                                         *
***************************************************************************
*                                                                         *
*  AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS. AUTODESK   *
*  SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF MERCHANTABILITY OR      *
*  FITNESS FOR A PARTICULAR USE.  AUTODESK, INC. DOES NOT WARRANT THAT    *
*  THE OPERATION OF THE PROGRAM WILL BE UNINTERRUPTED OR ERROR FREE.      *
*                                                                         *
***************************************************************************
*                                                                         *
*                       RESTRICTED RIGHTS LEGEND                          *
*                                                                         *
*  Use, duplication, or disclosure by the U.S. Government is subject      *
*  to restrictions set forth in FAR 52.227-19 (Commercial Computer        *
*  Software - Restricted Rights) and DFAR 252.227-7013 (c) (1) (ii)       *
*  (Rights in Technical Data and Computer Software), as applicable.       *
***************************************************************************

   Modification History:

   Notes and restrictions on use:
       This module provides a geometry library with functions for manipulating 
       and creating geometrical entities.  This library can be expanded any
       time later when there is a need for more functions. 
       
***************************************************************************/
/**************************************************************************/
/*  STANDARD C INCLUDE FILES  */
/**************************************************************************/

#include  <stdio.h>
#include  <math.h> 
#include  <assert.h> 
#include  <string.h>
#include  "adslib.h"
#include  "sdk.h"
#include  "gl.h"

 
/**************************************************************************/
/*  GLOBAL DEFINES  */
/**************************************************************************/
#define X        0
#define Y        1
#define Z        2     
#define LLL_STORE_END   'E'
#define LLL_STORE_HEAD  'H'
#if !defined(max)
#define max(a,b)  ((a)>(b) ? (a) : (b))
#endif
#define TOL      1e-10
#define TINY     1e-20
#define ARBBOUND 0.015625
 
 
/**************************************************************************/
/*  GLOBAL TYPEDEFS  */
/**************************************************************************/

 
/**************************************************************************/
/*  GLOBAL VARIABLES   */
/**************************************************************************/

extern char *lll_alloc_node();
extern int lll_store_node();
pvtype    *pvhead = NULL;
pltype    *plhead = NULL;


/**************************************************************************/
/*.doc gl_ac2wc(m, ndir)*/
/*+
   This function creates the transformation matrix from arbitrary coord.
   system to World Coordinate System.
-*/
/**************************************************************************/
int
/* FCN */gl_ac2wc(m, ndir)
txform     m;                          /* (O) transformation matrix */
ads_point  ndir;                       /* (I) vector in normal direction */
{
register int     i;
ads_point  xdir, ydir;

   gl_mat_ident(m);
   if (fabs(ndir[X])>0.0 || fabs(ndir[Y])>0.0) { 
      if (gl_arbaxis(xdir,ndir)==BAD) return BAD;
      if (gl_vec_cross(ydir,ndir,xdir)==BAD) return BAD;
      if (gl_vec_unit(ydir,ydir)==BAD) return BAD;
      for (i=0; i<3; i++) {
         m[i][0] = xdir[i];
         m[i][1] = ydir[i];
         m[i][2] = ndir[i];
      }
   }
   return GOOD;

/*  END GL_AC2WC FUNCTION  */
}


/**************************************************************************/
/*.doc gl_ang_txf2d(m)*/
/*+
   This function returns the rotation angle from a transformation matrix
   which represents a 2D rigid body motion.
-*/
/**************************************************************************/
ads_real
/* FCN */gl_ang_txf2d(m)
txform     m;                          /* (I) transformation matrix */
{
ads_real   a, b, theta;
 
   a = m[0][0];  b = m[1][0];
   theta = acos(a);
   if (b<0.0) theta = 2.0*PI - theta;
   return theta;

/*  END GL_ANG_TXF2D FUNCTION  */
}


/**************************************************************************/
/*.doc gl_arbaxis(oaxis, iaxis)*/
/*+
   This function creates an arbitrary axis perpendicular to a given
   axis using the AutoCAD arbitrary axis algorithm.
-*/
/**************************************************************************/
int
/* FCN */gl_arbaxis(oaxis, iaxis)
ads_point  oaxis;                      /* (O) output axis */
ads_point  iaxis;                      /* (I) input axis */
{
register int   i;
ads_point  waxis;

   for (i=0; i<3; i++) waxis[i] = 0.0;
   if (fabs(iaxis[X])<ARBBOUND && fabs(iaxis[Y])<ARBBOUND) {
      waxis[Y] = 1.0;
      if (gl_vec_cross(oaxis,waxis,iaxis)==BAD) return BAD;
   }
   else {
       waxis[Z] = 1.0;
       if (gl_vec_cross(oaxis,waxis,iaxis)==BAD) return BAD;
   }
   if (gl_vec_unit(oaxis,oaxis)==BAD) return BAD;
   return GOOD;

/*  END GL_ARBAXIS FUNCTION  */
}


/**************************************************************************/
/*.doc gl_bul2cir(circle, p1, p2, bf)*/
/*+
   This function converts two vertices into a circle with a bulge factor.
-*/
/**************************************************************************/
int
/* FCN */gl_bul2cir(circle, p1, p2, bf)
ctype      *circle;                    /* (O) circle information */
ads_point  p1, p2;                     /* (I) two points */
ads_real   bf;                         /* (I) bulge factor */
{
ads_real   cotbce;

   cotbce = 0.5*((1.0/bf)-bf);
   circle->pt[X] = 0.5*(p1[X]+p2[X]-(p2[Y]-p1[Y])*cotbce);
   circle->pt[Y] = 0.5*(p1[Y]+p2[Y]+(p2[X]-p1[X])*cotbce);
   circle->pt[Z] = 0.0;
   circle->radius = gl_dpp(p1,circle->pt);
   return GOOD;
 
/*  END GL_BUL2CIR FUNCTION  */
}


/**************************************************************************/
/*.doc gl_cir2eli(conic, mat, circle)*/
/*+
   This function is used to convert a circle in a coordinate system to
   an ellipse when the circle is projected to the World Coordinate System.
-*/
/**************************************************************************/
int
/* FCN */gl_cir2eli(conic, mat, circle)
cotype        *conic;                          /* (O) pointer to the conic data */
txform        mat;                             /* (I) transformation matrix */
ctype         *circle;                         /* (I) pointer to the circle data */
{
ads_real  a11, a12, a13,
          a21, a22, a23,
          a31, a32, a33,
          x0, y0, z0,
          b1, b2, b3, b4;

   a11 = mat[0][0];
   a21 = mat[1][0];
   a31 = mat[2][0];
   a12 = mat[0][1];
   a22 = mat[1][1];
   a32 = mat[2][1];
   a13 = mat[0][2];
   a23 = mat[1][2];
   a33 = mat[2][2];
   x0 = mat[0][3];
   y0 = mat[1][3];
   z0 = mat[2][3];
   b1 = (a13*a13+a23*a23)/(a33*a33);
   b2 = 2.0*(a11*a13+a21*a23)/a33;
   b3 = 2.0*(a12*a13+a22*a23)/a33;
   b4 = 2.0*(a13*x0+a23*y0)/a33;
   conic->a = a11*a11+a21*a21+b1*a31*a31-b2*a31;
   conic->b = 2.0*(a11*a12+a21*a22)+2.0*b1*a31*a32-b2*a32-b3*a31;
   conic->c = a12*a12+a22*a22+b1*a32*a32-b3*a32;
   conic->d = 2.0*(a11*x0+a21*y0)+2.0*b1*a31*z0-b2*z0-b4*a31;
   conic->e = 2.0*(a12*x0+a22*y0)+2.0*b1*a32*z0-b3*z0-b4*a32;
   conic->f = x0*x0+y0*y0+b1*z0*z0-b4*z0-circle->radius*circle->radius;
   return GOOD;
 
/*  END GL_CIR2ELI FUNCTION  */
}


/**************************************************************************/
/*.doc gl_dpl(pt, line)*/
/*+
   This function is used to calculate the distance between a point and a
   line.
-*/
/**************************************************************************/
ads_real
/* FCN */gl_dpl(pt, line)
ads_point  pt;                         /* (I) a point */
ltype      *line;                      /* (I) line equation */
{
ads_real   dx, dy, dz,t1,t2;

   dx = pt[X]-line->pt1[X];
   dy = pt[Y]-line->pt1[Y];
   dz = pt[Z]-line->pt1[Z];
   t1 = dx*line->cos[X]+dy*line->cos[Y]+dz*line->cos[Z];
   t2 = dx*dx+dy*dy+dz*dz-t1*t1;
   return sqrt(t2);
 
/*  END GL_DPL FUNCTION  */
}


/**************************************************************************/
/*.doc gl_dpp(p1, p2)*/
/*+
   This function is used to calculate the distance between two points.
-*/
/**************************************************************************/
ads_real
/* FCN */gl_dpp(p1, p2)
ads_point  p1, p2;                     /* (I) two points */
{
ads_point   pd;

   gl_vec_sub(pd,p1,p2);
   return gl_vec_abs(pd);  
 
/*  END GL_DPP FUNCTION  */
}


/**************************************************************************/
/*.doc gl_dist_ptpln(dist, pt, pt1, pt2, pt3)*/
/*+
   This function is used to calculate the distance between a point and a
   plane which is made up of three points pt1, pt2 and pt3.
-*/
/**************************************************************************/
int
/* FCN */gl_dist_ptpln(dist, pt, pt1, pt2, pt3)
ads_real   *dist;                      /* (O) distance */
ads_point  pt;                         /* (I) a point */
ads_point  pt1, pt2, pt3;              /* (I) three points to define a plane */
{
ads_point  pqpt;

   if (gl_pt_pq2pln(pqpt,pt,pt1,pt2,pt3)== BAD) return BAD;
   *dist = gl_dpp(pqpt,pt);
   return GOOD;
 
/*  END GL_DIST_PTPLN FUNCTION  */
}


/**************************************************************************/
/*.doc gl_fixang(angle)*/
/*+
   This function is used to reduce an angle to the range 0<=angle<2*PI.
-*/
/**************************************************************************/
ads_real
/* FCN */gl_fixang(angle)
ads_real   angle;                      /* (I) angle */
{
ads_real   iang;

   if (angle>=0.0) iang = floor((angle/(2.0*PI)));
   else iang = ceil((angle/(2.0*PI)-1.0)); 
   return angle-(2.0*PI*iang);
 
/*  END GL_FIXANG FUNCTION  */
}


/**************************************************************************/
/*.doc gl_intlc_pln(pts, line, circle)*/
/*+
   This function is used to calculate the intersection point(s) of a line
   and a circle on a plane (2D).
-*/
/**************************************************************************/
int
/* FCN */gl_intlc_pln(pts, line, circle)
ads_point  pts[2];                     /* (O) intersection points */
ltype      *line;                      /* (I) line data */
ctype      *circle;                    /* (I) circle data */
{
ads_real   t1, t2, a, b, c, d, e, f, 
           alpha, beta, gamma,
           delta, root1, root2,
           sqb, ag4;
int        m, fy, nroots;

   for (m=0; m<2; m++) {
      pts[m][X]=0.0;
      pts[m][Y]=0.0;
      pts[m][Z]=0.0;
   }
   a = line->g;
   b = line->h;
   c = line->i;
   t1 = a*a;
   t2 = b*b;
   d = -2.0*circle->pt[X];
   e = -2.0*circle->pt[Y];
   f = d*d/4.0 + e*e/4.0 - circle->radius*circle->radius;
   if (0!=(fy = (fabs(a)>fabs(b)))) {
      alpha = t2/t1 + 1.0;
      beta = -d*b/a + 2.0*c*b/t1 + e;
      gamma = c*c/t1 - d*c/a + f;
   }
   else {
      alpha = t1/t2 + 1.0;
      beta = -e*a/b + 2.0*a*c/t2 + d;
      gamma = c*c/t2 -e*c/b + f;
   }
   delta = (sqb=beta*beta) - (ag4=4.0*alpha*gamma);
   if (fabs(delta)<(max(fabs(sqb),fabs(ag4))*TOL)) {
      root1 = -beta/(2.0*alpha);
      nroots = 1;
   }
   else
      if (delta<0.0) return 0;
      else {
         delta = sqrt(delta);
         nroots = 2;
         root1 = -(beta+delta)/(2.0*alpha);
         root2 = root1+delta/alpha;
      }
   if (fy) {
      pts[0][Y] = root1;
      pts[0][X] = -(b*pts[0][Y]+c)/a;
      if (nroots==2) {
         pts[1][Y] = root2;
         pts[1][X] = -(b*pts[1][Y]+c)/a;
      }
   }
   else {
      pts[0][X] = root1;
      pts[0][Y] = -(a*pts[0][X]+c)/b;
      if (nroots==2) {
         pts[1][X] = root2;
         pts[1][Y] = -(a*pts[1][X]+c)/b;
      }
   }
   return nroots;
 
/*  END GL_INTLC_PLN FUNCTION  */
}


/**************************************************************************/
/*.doc gl_intcc_pln(pts, circ1, circ2)*/
/*+
   This function is used to calculate the intersection point(s) of two
   circles on a plane and returns the number of intersection points.
   If the two circles coincide, the value -1 will be returned.
-*/
/**************************************************************************/
int
/* FCN */gl_intcc_pln(pts, circ1, circ2)
ads_point  pts[2];                     /* (O) intersection points */
ctype      *circ1, *circ2;             /* (I) circles data */
{
ads_real   d1, d2, e1, e2, f1, f2;
ltype      line;
int        n; 
 
   d1 = -2.0*circ1->pt[X];
   e1 = -2.0*circ1->pt[Y];
   f1 = d1*d1/4.0 + e1*e1/4.0 - circ1->radius*circ1->radius;
   d2 = -2.0*circ2->pt[X];
   e2 = -2.0*circ2->pt[Y];
   f2 = d2*d2/4.0 + e2*e2/4.0 - circ2->radius*circ2->radius;
   line.g = d1-d2;
   line.h = e1-e2;
   line.i = f1-f2;
   if (line.g==0 && line.h==0) return -1;
   n = gl_intlc_pln(pts,&line,circ1); 
   return n;
 
/*  END GL_INTCC_PLN FUNCTION  */
}


/**************************************************************************/
/*.doc gl_intll_coef(pt, line1, line2)*/
/*+
   This function is used to calculate the intersection point of two
   lines on the same plane in terms of their coefficients.
-*/
/**************************************************************************/
int
/* FCN */gl_intll_coef(pt, line1, line2)
ads_point  pt;                         /* (O) a point */
ltype      *line1, *line2;             /* (I) line data */
{
ads_real   t;

   t = line1->g*line2->h - line2->g*line1->h;
   if (fabs(t)<=TOL) return BAD;
   pt[0] = (line1->h*line2->i - line2->h*line1->i)/t;
   pt[1] = (line1->i*line2->g - line2->i*line1->g)/t;
   pt[2] = 0.0;
   return GOOD;
 
/*  END GL_INTLL_COEF FUNCTION  */
}


/**************************************************************************/
/*.doc gl_int_lco(pts, line, conic)*/
/*+
   This function is used to calculate the intersection point(s) of a line
   and a conic.
-*/
/**************************************************************************/
int
/* FCN */gl_int_lco(pts, line, conic)
ads_point  pts[2];                     /* (O) intersection points */
ltype      *line;                      /* (I) line data */
cotype     *conic;                     /* (I) conic data */
{
ads_real   t1, t2, a, b, c, d, e, f, 
           g, h, i,
           alpha, beta, gamma,
           delta, root1, root2,
           sqb, ag4;
int        m, n, fy, nroots;

   for (m=0; m<2; m++) 
      for (n=0; n<3; n++) pts[m][n] = 0.0;
   g = line->g;
   h = line->h;
   i = line->i;
   t1 = g*g;
   t2 = h*h;
   a = conic->a;
   b = conic->b;
   c = conic->c;
   d = conic->d;
   e = conic->e;
   f = conic->f;
   if (0!=(fy = (fabs(g)>fabs(h)))) {
      alpha = a*t2/t1 - b*h/g + c;
      beta = -(b*i+d*h)/g + 2.0*i*h*a/t1 + e;
      gamma = a*i*i/t1 - d*i/g + f;
   }
   else {
      alpha = c*t1/t2 - b*g/h + a;
      beta = -(b*i+e*g)/h + 2.0*g*i*c/t2 + d;
      gamma = c*i*i/t2 -e*i/h + f;
   }
   delta = (sqb=beta*beta) - (ag4=4.0*alpha*gamma);
   if (fabs(delta)<(max(fabs(sqb),fabs(ag4))*TOL)) {
      root1 = -beta/(2.0*alpha);
      nroots = 1;
   }
   else
      if (delta<0.0) return 0;
      else {
         delta = sqrt(delta);
         nroots = 2;
         root1 = -(beta+delta)/(2.0*alpha);
         root2 = root1+delta/alpha;
      }
   if (fy) {
      pts[0][Y] = root1;
      pts[0][X] = -(h*pts[0][Y]+i)/g;
      if (nroots==2) {
         pts[1][Y] = root2;
         pts[1][X] = -(h*pts[1][Y]+i)/g;
      }
   }
   else {
      pts[0][X] = root1;
      pts[0][Y] = -(g*pts[0][X]+i)/h;
      if (nroots==2) {
         pts[1][X] = root2;
         pts[1][Y] = -(g*pts[1][X]+i)/h;
      }
   }
   return nroots;
 
/*  END GL_INT_LCO FUNCTION  */
}


/**************************************************************************/
/*.doc gl_line(line, p1, p2)*/
/*+
   This function is used to set up line equation from two points.  It returns
   value of 1 for successful setup, otherwise 0 is returned.
-*/
/**************************************************************************/
int
/* FCN */gl_line(line, p1, p2)
ltype      *line;                      /* (O) line equation */
ads_point  p1, p2;                     /* (I) two points */
{
ads_point  pd;
ads_real   dist;
register int  i;

   gl_vec_sub(pd,p2,p1);
   dist = gl_vec_abs(pd);
   if (dist==0) return BAD;
   for (i=0; i<3; i++) {
      line->pt1[i] = p1[i];
      line->pt2[i] = p2[i];
      line->cos[i] = pd[i]/dist;
   }
   return GOOD;
 
/*  END GL_LINE FUNCTION  */
}


/**************************************************************************/
/*.doc gl_line_coef(line, p1, p2)*/
/*+
   This function is used to set up the coefficients of line equation from
   two points.  It returns GOOD for successful setup, otherwise BAD is
   returned.
-*/
/**************************************************************************/
int
/* FCN */gl_line_coef(line, p1, p2)
ltype      *line;                      /* (O) line data */
ads_point  p1, p2;                     /* (I) two points */
{

   line->g = p2[Y]-p1[Y];
   line->h = p1[X]-p2[X];
   if (line->g==0 && line->h==0) return BAD;
   line->i = -(line->g*p1[X]+line->h*p1[Y]);
   return GOOD;
 
/*  END GL_LINE_COEF FUNCTION  */
}


/**************************************************************************/
/*.doc gl_ludcmp(m, indx, d)*/
/*+
   This function replaces the matrix m by the LU decomposition of a rowwise
   permutation of itself.  indx is an output vector which records the 
   row permutation effected by the partial pivoting.  d is output as 1 or -1
   depending on whether the number of row interchanges was even or odd.
-*/
/**************************************************************************/
int
/* FCN */gl_ludcmp(m, indx, d)
matNxN     m;                          /* (I/O) input matrix and its decomp.*/
int        indx[ORDER];                /* (O) permutation vector */ 
ads_real   *d;                         /* (O) 1 or -1 */
{
int        i, imax, j, k;
ads_real   big, dum, sum, temp;
ads_real   vv[ORDER];

   *d = 1.0;
   for (i=0; i<ORDER; i++) {
      big = 0.0;
      for (j=0; j<ORDER; j++)
         if ((temp=fabs(m[i][j]))>big) big = temp;
      if (big==0.0) return BAD;
      vv[i] = 1.0/big;
   }
   for (j=0; j<ORDER; j++) {
      for (i=0; i<j; i++) {
         sum = m[i][j];
         for (k=0; k<i; k++) sum -= m[i][k]*m[k][j];
         m[i][j] = sum;
      }
      big = 0.0;
      for (i=j; i<ORDER; i++) {
         sum = m[i][j];
         for (k=0; k<j; k++) sum -= m[i][k]*m[k][j];
         m[i][j] = sum;
         if ((dum=vv[i]*fabs(sum))>=big) {
            big = dum;
            imax = i;
         }
      }
      if (j!=imax) {
         for (k=0; k<ORDER; k++) {
            dum = m[imax][k];
            m[imax][k] = m[j][k];
            m[j][k] = dum;
         }
         *d = -(*d);
         vv[imax] = vv[j];
      }  
      indx[j] = imax;
      if (m[j][j]==0.0) m[j][j] = TINY;
      if (j!=ORDER-1) {
         dum = 1.0/(m[j][j]);
         for (i=j+1; i<ORDER; i++) m[i][j] *= dum;
      }
   }  
   return GOOD;
 
/*  END GL_LUDCMP FUNCTION  */
}


/**************************************************************************/
/*.doc gl_lubksb(m, indx, b)*/
/*+
   This function solves the set of n (ORDER) linear equations MX=B.  The 
   matrix m is not as the matrix M but rather as its LU decomposition,
   returned by gl_ludcmp.  indx is the permutation vector also returned by
   gl_ludcmp.  b is the right-hand side vector B and returns the solution
   vector X.
-*/
/**************************************************************************/
void
/* FCN */gl_lubksb(m, indx, b)
matNxN     m;                          /* (I) LU decomposition coef. matrix */
int        indx[ORDER];                /* (I) permutation vector */ 
ads_real   b[ORDER];                   /* (I) RHS vector and solution vector */ 
{
int        i, id=0, ii = 0, ip, j;
ads_real   sum;

   for (i=0; i<ORDER; i++) {
      ip = indx[i];
      sum = b[ip];
      b[ip] = b[i];
      if (id)
         for (j=ii; j<=i-1; j++) sum -= m[i][j]*b[j];
      else
         if (sum) {
            ii = i;
            id = 1;
         }
      b[i] = sum;
   }
   for (i=ORDER-1; i>=0; i--) {
      sum = b[i];
      for (j=i+1; j<ORDER; j++) sum -= m[i][j]*b[j];
      b[i] = sum/m[i][i];
   }
   return;

/*  END GL_LUBKSB FUNCTION  */
}


/**************************************************************************/
/*.doc gl_matXmat(ma, mb, mc)*/
/*+
   This function multiplies two 4X4 matrices mb and mc and outputs the  
   4X4 matrix ma.
-*/
/**************************************************************************/
void
/* FCN */gl_matXmat(ma, mb, mc)
txform     ma;                         /* (O) 4X4 output matrix */
txform     mb, mc;                     /* (I) 4X4 input matrices */
{
register int i, j, k;
ads_real   sum;

   for (i=0; i<4; i++) 
      for (j=0; j<4; j++) {
         sum = 0.0;
         for (k=0; k<4; k++) sum = sum + mb[i][k]*mc[k][j];
         ma[i][j] = sum;
      }
   return;

/*  END GL_MATXMAT FUNCTION  */
}


/**************************************************************************/
/*.doc gl_mat_det(m)*/
/*+
   This function calculates the determinant of a matrix m and returns its
   value.
-*/
/**************************************************************************/
ads_real
/* FCN */gl_mat_det(m)
matNxN     m;                          /* (I) input matrix */
{
register int   j;
int        indx[ORDER];
ads_real   d;

   if (gl_ludcmp(m,indx,&d)==BAD) return 0.0;
   for (j=0; j<ORDER; j++) d *= m[j][j];
   return d;
 
/*  END GL_MAT_DET FUNCTION  */
}


/**************************************************************************/
/*.doc gl_mat_inv(invm, m)*/
/*+
   This function finds the inverse matrix of the matrix m.
-*/
/**************************************************************************/
int
/* FCN */gl_mat_inv(invm, m)
matNxN     invm;                       /* (O) inverse matrix */
matNxN     m;                          /* (I) input matrix */
{
register int      i, j;
int        indx[ORDER];
ads_real   d;
ads_real   col[ORDER];
matNxN     tm;

   for (i=0; i<ORDER; i++)
      for (j=0; j<ORDER; j++) tm[i][j] = m[i][j];
   if (gl_ludcmp(m,indx,&d)==BAD) return BAD;
   for (j=0; j<ORDER; j++) {
      for (i=0; i<ORDER; i++) col[i] = 0.0;
      col[j] = 1.0;
      gl_lubksb(m,indx,col);
      for (i=0; i<ORDER; i++) invm[i][j] = col[i];
   }
   for (i=0; i<ORDER; i++)
      for (j=0; j<ORDER; j++) m[i][j] = tm[i][j];
   return GOOD;

/*  END GL_MAT_INV FUNCTION  */
}


/**************************************************************************/
/*.doc gl_mat_ident(m)*/
/*+
   This function creates a 4X4 identity matrix m.
-*/
/**************************************************************************/
void
/* FCN */gl_mat_ident(m)
txform     m;                      /* (O) 4X4 identity matrix */
{
register int  i, j;

   for (i=0; i<ORDER; i++)
      for (j=0; j<ORDER; j++)
         m[i][j] = i==j ? 1.0 : 0.0;
   return;

/*  END GL_MAT_IDENT FUNCTION  */
}


/**************************************************************************/
/*.doc gl_mat_trans(m, x, y, z)*/
/*+
   This function updates a transformation matrix which corresponds to a
   translation.
-*/
/**************************************************************************/
void
/* FCN */gl_mat_trans(m, x, y, z)
txform     m;                          /* (O) transformation matrix */
ads_real   x, y, z;                    /* (I) translation values */
{
   gl_mat_ident(m);
   m[0][3] = x;
   m[1][3] = y;
   m[2][3] = z;
   return;

/*  END GL_MAT_TRANS FUNCTION  */
}


/**************************************************************************/
/*.doc gl_mat_rotx(m, theta)*/
/*+
   This function updates a transformation matrix which corresponds to a
   rotation about the X-axis.
-*/
/**************************************************************************/
void
/* FCN */gl_mat_rotx(m, theta)
txform     m;                          /* (O) transformation matrix */
ads_real   theta;                      /* (I) rotation angle about X-axis */
{
   gl_mat_ident(m);
   m[1][1] = m[2][2] = cos(theta);
   m[1][2] = -(m[2][1] = sin(theta));
   return;

/*  END GL_MAT_ROTX FUNCTION  */
}


/**************************************************************************/
/*.doc gl_mat_roty(m, theta)*/
/*+
   This function updates a transformation matrix which corresponds to a
   rotation about the Y-axis.
-*/
/**************************************************************************/
void
/* FCN */gl_mat_roty(m, theta)
txform     m;                          /* (O) transformation matrix */
ads_real   theta;                      /* (I) rotation angle about Y-axis */
{
   gl_mat_ident(m);
   m[0][0] = m[2][2] = cos(theta);
   m[2][0] = -(m[0][2] = sin(theta));
   return;

/*  END GL_MAT_ROTY FUNCTION  */
}


/**************************************************************************/
/*.doc gl_mat_rotz(m, theta)*/
/*+
   This function updates a transformation matrix which corresponds to a
   rotation about the Z-axis.
-*/
/**************************************************************************/
void
/* FCN */gl_mat_rotz(m, theta)
txform     m;                          /* (O) transformation matrix */
ads_real   theta;                      /* (I) rotation angle about Z-axis */
{
   gl_mat_ident(m);
   m[0][0] = m[1][1] = cos(theta);
   m[0][1] = -(m[1][0] = sin(theta));
   return;

/*  END GL_MAT_ROTZ FUNCTION  */
}


/**************************************************************************/
/*.doc gl_mat_multi(ap, bp, m, disp)*/
/*+
   This function multiplies a point or a displacement by a matrix in order
   to perform a transformation.
-*/
/**************************************************************************/
void
/* FCN */gl_mat_multi(ap, bp, m, disp)
ads_point  ap;                         /* (O) transformed point or disp. */
ads_point  bp;                         /* (I) point/disp. to be transformed */
txform     m;                          /* (I) transformation matrix */
int        disp;                       /* (I) flag indicating point or disp. */
{
register int i;
ads_point  sum;

   for (i=0; i<3; i++)
      sum[i] = m[i][0]*bp[X]+m[i][1]*bp[Y]+m[i][2]*bp[Z]
              +(disp ? 0.0 : m[i][3]);
   for (i=0; i<3; i++) ap[i] = sum[i];
   return;

/*  END GL_MAT_MULTI FUNCTION  */
}


/**************************************************************************/
/*.doc gl_mat_scal(ma, b, mc)*/
/*+
   This function multiplies a scalar b to the 4X4 matrix mc and outputs the 
   4X4 matrix ma.
-*/
/**************************************************************************/
void
/* FCN */gl_mat_scal(ma, b, mc)
txform     ma;                         /* (O) 4X4 output matrix */
ads_real   b;                          /* (I) a scalar */
txform     mc;                         /* (I) 4X4 input matrix */
{
register int   i, j;

   for (i=0; i<4; i++) 
      for (j=0; j<4; j++) ma[i][j] = b * mc[i][j];
   return;

/*  END GL_MAT_SCAL FUNCTION  */
}


/**************************************************************************/
/*.doc gl_pln_pts(pln, p1, p2, p3)*/
/*+
   This function finds the unit normal to a plane from three points which
   lie on the plane.
-*/
/**************************************************************************/
int
/* FCN */gl_pln_pts(pln, p1, p2, p3)
ads_point  pln;                    /* (O) normal vector to the plane */
ads_point  p1, p2, p3;             /* (I) input points */
{
ads_point  v1, v2;

   gl_vec_sub(v1,p2,p1);
   gl_vec_sub(v2,p3,p1);
   return gl_pln_vec(pln,v1,v2);
  
/*  END GL_PLN_PTS FUNCTION  */
}


/**************************************************************************/
/*.doc gl_pln_vec(pln, v1, v2)*/
/*+
   This function finds the unit norm to a plane from two vectors.
-*/
/**************************************************************************/
int
/* FCN */gl_pln_vec(pln, v1, v2)
ads_point  pln;                    /* (O) normal vector to the plane */
ads_point  v1, v2;                 /* (I) input vectors */
{

   if (gl_vec_cross(pln,v1,v2)==BAD) return BAD;
   return gl_vec_unit(pln,pln);
  
/*  END GL_PLN_VEC FUNCTION  */
}


/**************************************************************************/
/*.doc gl_ptseg(line1, pt)*/
/*+
   This function is used to test a point being on a line segment or not.
-*/
/**************************************************************************/
int
/* FCN */gl_ptseg(line1, pt)
ltype      *line1;                   /* (I) one of the intersection lines */
ads_point  pt;                       /* (I) an intersection point */
{
ads_real   a,b,c;

   a = pt[X] - line1->pt1[X];
   b = line1->pt2[X]-line1->pt1[X];
   c = a/b;
   if (c <= 1.0 && c >= 0.0) return TRUE;
   else return FALSE;
 
/*  END GL_PTSEG FUNCTION  */
}


/**************************************************************************/
/*.doc gl_pts_cross(nv, ap, bp, cp)*/
/*+
   This function is used to compute the cross product from three points.
   The first point ap is the origin and the vector from the second point
   to the first point is the X-axis, i.e., nv = (bp-ap) X (cp-ap).
-*/
/**************************************************************************/
int
/* FCN */gl_pts_cross(nv, ap, bp, cp)
ads_point  nv;                     /* (O) result vector */
ads_point  ap, bp, cp;             /* (I) points to compute cross product */
{
ads_point  v1, v2;

   gl_vec_sub(v1,bp,ap);
   gl_vec_sub(v2,cp,ap);
   if (gl_vec_cross(nv,v1,v2)==BAD) return BAD;
   return GOOD;

/*  END GL_PTS_CROSS FUNCTION  */
}


/**************************************************************************/
/*.doc gl_pt_pq2l2d(pqpt, pt, line)*/
/*+
   This function is used to calculate the perpendicular from a point to 
   a line segment.
-*/
/**************************************************************************/
int
/* FCN */gl_pt_pq2l2d(pqpt, pt, line)
ads_point  pqpt;                       /* (O) the perpendinular */
ads_point  pt;                         /* (I) a point */
ltype      *line;                      /* (I) line information */
{
ads_point  vec;
ads_real   dsq, xsq, ysq;

   gl_vec_sub(vec,line->pt2,line->pt1);
   if ((dsq=(xsq=vec[X]*vec[X])+(ysq=vec[Y]*vec[Y]))>0.0) {
      pqpt[X] = (line->pt1[X]*ysq+pt[X]*xsq+(pt[Y]-line->pt1[Y])
                *vec[Y]*vec[X])/dsq;
      pqpt[Y] = (line->pt1[Y]*xsq+pt[Y]*ysq+(pt[X]-line->pt1[X])
                *vec[X]*vec[Y])/dsq;
      pqpt[Z] = 0.0;
      return GOOD;
   }
   else return BAD;
 
/*  END GL_PT_PQ2L2D FUNCTION  */
}


/**************************************************************************/
/*.doc gl_pt_pq2l3d(pqpt, pt, line)*/
/*+
   This function is used to calculate the perpendicular from a point to
   a line segment in 3D space.
-*/
/**************************************************************************/
int
/* FCN */gl_pt_pq2l3d(pqpt, pt, line)
ads_point     pqpt;                              /* (O) perpendicular point */
ads_point     pt;                                /* (I) input point */
ltype         *line;                             /* (I) pointer to the line data */
{
ads_point     nvec, spt1, spt2;
int           disp, i;
txform        invm, m;

   if (gl_pln_pts(nvec,line->pt1,line->pt2,pt)==GOOD) {
      if (gl_ac2wc(m,nvec)==BAD) return BAD;
      for (i=0; i<3; i++) m[i][3] = line->pt1[i];
      if (gl_mat_inv(invm,m)==BAD) return BAD;
      disp = 0;
      setpt(spt1,line->pt1);
      setpt(spt2,line->pt2);
      gl_mat_multi(line->pt1,line->pt1,invm,disp);
      gl_mat_multi(line->pt2,line->pt2,invm,disp);
      gl_mat_multi(pt,pt,invm,disp);
      gl_pt_pq2l2d(pqpt,pt,line);
      gl_mat_multi(pqpt,pqpt,m,disp);
      setpt(line->pt1,spt1);
      setpt(line->pt2,spt2);
   }
   else {
      if (gl_vec_eqv(line->pt1,line->pt2)) {
         pt[X] = line->pt1[X];
         pt[Y] = line->pt1[Y];
      }
      setpt(pqpt,pt);
   }
   return GOOD;
 
/*  END GL_PT_PQ2L3D FUNCTION  */
}


/**************************************************************************/
/*.doc gl_pt_pq2pln(pqpt, pt, pt1, pt2, pt3)*/
/*+
   This function is used to calculate the perpendicular from a point to
   a plane which is made up of three points pt1, pt2, and pt3 in 3D space.
-*/
/**************************************************************************/
int
/* FCN */gl_pt_pq2pln(pqpt, pt, pt1, pt2, pt3)
ads_point     pqpt;                     /* (O) perpendicular point */
ads_point     pt;                       /* (I) input point */
ads_point     pt1, pt2, pt3;            /* (I) three points to define a plane */
{
ads_point     nvec, spt;
int           disp, i;
txform        invm, m;

   if (gl_pln_pts(nvec,pt1,pt2,pt3)==BAD) return BAD;
   if (gl_ac2wc(m,nvec)==BAD) return BAD;
   for (i=0; i<3; i++) m[i][3] = pt1[i];
   if (gl_mat_inv(invm,m)==BAD) return BAD;
   disp = 0;
   setpt(spt,pt);
   gl_mat_multi(pt,pt,invm,disp);
   setpt(pqpt,pt);
   pqpt[Z] = 0.0;
   gl_mat_multi(pqpt,pqpt,m,disp);
   setpt(pt,spt);
   return GOOD;
 
/*  END GL_PT_PQ2PLN FUNCTION  */
}


/**************************************************************************/
/*.doc gl_pt_oncirc(pt, circle, angle)*/
/*+
   This function is used to find a point on a circle given an angle.
-*/
/**************************************************************************/
void
/* FCN */gl_pt_oncirc(pt, circle, angle)
ads_point  pt;                         /* (O) point */
ctype      *circle;                    /* (I) circle data */
ads_real   angle;                      /* (I) angle */
{

   pt[X] = circle->radius*cos(angle)+circle->pt[X];
   pt[Y] = circle->radius*sin(angle)+circle->pt[Y];
   pt[Z] = 0.0;
   return;

/*  END GL_PT_ONCIRC FUNCTION  */
}


/**************************************************************************/
/*.doc gl_pt_nstpl(npt, pt, line)*/
/*+
   This function is used to find the nearest point on a line segment to a
   given point.
-*/
/**************************************************************************/
int
/* FCN */gl_pt_nstpl(npt, pt, line)
ads_point     npt;                               /* (O) nearest point */
ads_point     pt;                                /* (I) input point */
ltype         *line;                             /* (I) pointer to the line data */
{
ads_real      t;

   gl_pt_pq2l3d(npt,pt,line);
   if (gl_ptseg(line,npt)==FALSE) {
      t = (npt[0]-line->pt1[0]) / (line->pt2[0]-line->pt1[0]);
      if (t<0.0) {
         setpt(npt,line->pt1);
      }
      else {
         setpt(npt,line->pt2);
      }
   }
   return GOOD;
 
/*  END GL_PT_NSTPL FUNCTION  */
}


/**************************************************************************/
/*.doc gl_p2lines()*/ 
/*+
   This function creates a linked list to hold the lines derived from 
   vertices.
-*/
/**************************************************************************/
int
/* FCN */gl_p2lines()
{
ads_point     vtx1,vtx2;
ads_real      bfac;
pltype        *plnode;
pvtype        *pvnode; 

   pvnode = pvhead;
   vtx1[X] = pvnode->vtx[X];
   vtx1[Y] = pvnode->vtx[Y];
   vtx1[Z] = pvnode->vtx[Z];
   bfac = pvnode->bulge;
   for (pvnode=pvnode->next; pvnode!=NULL; pvnode=pvnode->next) {
      vtx2[X] = pvnode->vtx[X];
      vtx2[Y] = pvnode->vtx[Y];
      vtx2[Z] = pvnode->vtx[Z];
      if ((plnode=(pltype *)lll_alloc_node(sizeof(pltype)))==NULL)
         return BAD;
      gl_line(&plnode->line,vtx1,vtx2);
      plnode->bf = 0;
      if (bfac!=0.0) {
         gl_bul2cir(&plnode->circle,vtx1,vtx2,bfac);
         plnode->bf = bfac>0.0 ? 1 : -1;
      }
      plnode->next = NULL;
      if (lll_store_node(plnode,&plhead,LLL_STORE_END)==RETURN_ABNORMAL) {
         return BAD;
      }
      vtx1[X] = vtx2[X];
      vtx1[Y] = vtx2[Y];
      vtx1[Z] = vtx2[Z];
      bfac = pvnode->bulge;
   }
   return GOOD;

/*  END GL_P2LINES FUNCTION  */
}
 
 
/**************************************************************************/
/*.doc gl_txf_rotpt(m, pt, theta)*/
/*+
   This function generates a 4X4 transformation matrix that rotates an 
   object theta degrees about the given point on the X-Y plane.
-*/
/**************************************************************************/
void
/* FCN */gl_txf_rotpt(m, pt, theta)
txform     m;                          /* (O) 4X4 output matrix */
ads_point  pt;                         /* (I) given point */
ads_real   theta;                      /* (I) rotation angle */
{
txform     m1, m2, m3;

   gl_mat_trans(m1,-pt[0],-pt[1],-pt[2]);
   gl_mat_rotz(m2,theta);
   gl_matXmat(m3,m2,m1);
   gl_mat_trans(m1,pt[0],pt[1],pt[2]);
   gl_matXmat(m,m1,m3);
   return;

/*  END GL_TXF_ROTPT FUNCTION  */
}


/**************************************************************************/
/*.doc gl_vec_sub(ap, bp, cp)*/
/*+
   This function is used to subtract vector cp from vector bp and the result  
   is vector ap, i.e. ap = bp - cp.  The vector consists of 3 elements, X, Y,
   and Z.
-*/
/**************************************************************************/
void
/* FCN */gl_vec_sub(ap, bp, cp)
ads_point  ap;                    /* (O) result vector */
ads_point  bp, cp;                /* (I) 2 vectors to operate subtraction*/
{
   ap[X] = bp[X] - cp[X];
   ap[Y] = bp[Y] - cp[Y];
   ap[Z] = bp[Z] - cp[Z];
   return;

/*  END GL_VEC_SUB FUNCTION  */
}


/**************************************************************************/
/*.doc gl_vec_scal(ap, b, cp)*/
/*+
   This function is used to multiply a vector cp by a scalar b.  The result
   vector is ap, i.e. ap = b * cp.  The vector consists of 3 elements, X, Y,
   and Z.
-*/
/**************************************************************************/
void
/* FCN */gl_vec_scal(ap, b, cp)
ads_point  ap;                    /* (O) result vector */
ads_real   b;                     /* (I) multiplication scalar */
ads_point  cp;                    /* (I) vector to be multiplied*/
{
   ap[X] = b * cp[X];
   ap[Y] = b * cp[Y];
   ap[Z] = b * cp[Z];
   return;

/*  END GL_VEC_SCAL FUNCTION  */
}


/**************************************************************************/
/*.doc gl_vec_abs(ap)*/
/*+
   This function is used to calculate the norm of vector ap.
-*/
/**************************************************************************/
ads_real
/* FCN */gl_vec_abs(ap)
ads_point  ap;                     /* (I) vector to calculate norm */
{
   return sqrt(ap[X]*ap[X] + ap[Y]*ap[Y] + ap[Z]*ap[Z]); 
  
/*  END GL_VEC_ABS FUNCTION  */
}


/**************************************************************************/
/*.doc gl_vec_dot(ap, bp)*/
/*+
   This function is used to compute the dot product of two vectors.
-*/
/**************************************************************************/
ads_real
/* FCN */gl_vec_dot(ap, bp)
ads_point  ap, bp;                 /* (I) vectors to compute dot product */
{
   return (ap[X]*bp[X] + ap[Y]*bp[Y] + ap[Z]*bp[Z]); 
  
/*  END GL_VEC_DOT FUNCTION  */
}


/**************************************************************************/
/*.doc gl_vec_cross(ap, bp, cp)*/
/*+
   This function is used to compute the cross product of two vectors.
-*/
/**************************************************************************/
int
/* FCN */gl_vec_cross(ap, bp, cp)
ads_point  ap;                     /* (O) result vector */
ads_point  bp, cp;                 /* (I) vectors to compute cross product */
{
   if (ap==bp || ap==cp) return BAD;
   ap[X] = bp[Y]*cp[Z] - bp[Z]*cp[Y];
   ap[Y] = bp[Z]*cp[X] - bp[X]*cp[Z];
   ap[Z] = bp[X]*cp[Y] - bp[Y]*cp[X];
   return GOOD;

/*  END GL_VEC_CROSS FUNCTION  */
}


/**************************************************************************/
/*.doc gl_vec_add(ap, bp, cp)*/
/*+
   This function is used to add two vectors bp and cp.  The result vector 
   is ap.
-*/
/**************************************************************************/
void
/* FCN */gl_vec_add(ap, bp, cp)
ads_point  ap;                     /* (O) result vector */
ads_point  bp, cp;                 /* (I) vectors to be added */
{
   ap[X] = bp[X] + cp[X];
   ap[Y] = bp[Y] + cp[Y];
   ap[Z] = bp[Z] + cp[Z];
   return;

/*  END GL_VEC_ADD FUNCTION  */
}


/**************************************************************************/
/*.doc gl_vec_sqabs(ap)*/
/*+
   This function returns the square of the absolute value of a vector ap.
-*/
/**************************************************************************/
ads_real
/* FCN */gl_vec_sqabs(ap)
ads_point  ap;                     /* (I) vector */
{
   return (gl_vec_dot(ap,ap));

/*  END GL_VEC_SQABS FUNCTION  */
}


/**************************************************************************/
/*.doc gl_vec_unit(uv, ap)*/
/*+
   This function normalizes a vector ap and outputs its unit vector uv. 
-*/
/**************************************************************************/
int
/* FCN */gl_vec_unit(uv, ap)
ads_point  uv;                     /* (O) unit vector */
ads_point  ap;                     /* (I) input vector */
{
ads_real d;

   d = gl_vec_abs(ap);
   if (d <= TOL) return BAD;
   uv[X] = ap[X]/d;
   uv[Y] = ap[Y]/d;
   uv[Z] = ap[Z]/d;
   return GOOD;
  
/*  END GL_VEC_UNIT FUNCTION  */
}


/**************************************************************************/
/*.doc gl_vec_eqv(v1, v2)*/
/*+
   This function returns TRUE if two vectors are equal, otherwise, returns
   FALSE.
-*/
/**************************************************************************/
int
/* FCN */gl_vec_eqv(v1, v2)
ads_point  v1, v2;                           /* (I) input vectors */
{
ads_real   sqv, sqvd;
ads_point  vd;

   gl_vec_sub(vd,v1,v2);
   sqvd = gl_vec_sqabs(vd);
   sqv = max(gl_vec_sqabs(v1),gl_vec_sqabs(v2));
   if (sqv >= (TOL*TOL))
      sqvd /= sqv;
   return sqvd<(TOL*TOL);
  
/*  END GL_VEC_EQV FUNCTION  */
}


/**************************************************************************/
/*.doc gl_vec_ang(ang, va, vb, vc)*/
/*+
   This function finds the angle between two vectors va and vb while looking
   down the third vector vc.
-*/
/**************************************************************************/
int
/* FCN */gl_vec_ang(ang, va, vb, vc)
ads_real   *ang;                              /* (O) angle */
ads_point  va, vb, vc;                        /* (I) input vectors */
{
ads_real   cang, val;
ads_point  v1;

   cang = gl_vec_dot(va,vb)/(gl_vec_abs(va)*gl_vec_abs(vb));
   *ang = acos(cang);
   if (gl_vec_cross(v1,va,vb)==BAD) return BAD;
   val = gl_vec_dot(v1,vc);
   if (val==0.0) return BAD;
   if (val<0.0) *ang = -*ang;
   return GOOD;
  
/*  END GL_VEC_ANG FUNCTION  */
}


/**************************************************************************/
/*.doc gl_wc2ac(m, ndir)*/
/*+
   This function creates the transformation matrix from World Coordinate
   System to arbitrary coordinate system.
-*/
/**************************************************************************/
int
/* FCN */gl_wc2ac(m, ndir)
txform     m;                          /* (O) transformation matrix */
ads_point  ndir;                       /* (I) vector in normal direction */
{
int        i;
ads_point  xdir, ydir;

   gl_mat_ident(m);
   if (fabs(ndir[X])>0.0 || fabs(ndir[Y])>0.0) { 
      if (gl_arbaxis(xdir,ndir)==BAD) return BAD;
      if (gl_vec_cross(ydir,ndir,xdir)==BAD) return BAD;
      if (gl_vec_unit(ydir,ydir)==BAD) return BAD;
      for (i=0; i<3; i++) {
         m[0][i] = xdir[i];
         m[1][i] = ydir[i];
         m[2][i] = ndir[i];
      }
   }
   return GOOD;

/*  END GL_WC2AC FUNCTION  */
}
