/***************************************************************************
   Module Name:  lll      sdk_src_lll_lll.c

   Description:  General linked list processing library

   Author:  Tools and Technologies Group
            Autodesk, Inc.
            Software Developers' Kit
   
   (c) Copyright 1989,1990, 1991 by Autodesk Inc.
***************************************************************************
*  Permission to use, copy, modify, and distribute this software and its  *
*  documentation for the purpose of creating applications for AutoCAD, is *
*  hereby granted in accordance with the terms of the License Agreement   *
*  accompanying this product.                                             *
*                                                                         *
***************************************************************************
*                                                                         *
*  AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS.  AUTODESK  *
*  SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF MERCHANTABILITY OR      *
*  FITNESS FOR A PARTICULAR USE.  AUTODESK, INC. DOES NOT WARRANT THAT    *
*  THE OPERATION OF THE PROGRAM WILL BE UNINTERRUPTED OR ERROR FREE.      *
*                                                                         *
***************************************************************************
*                                                                         *
*                       RESTRICTED RIGHTS LEGEND                          *
*                                                                         *
*  Use, duplication, or disclosure by the U.S. Government is subject      *
*  to restrictions set forth in FAR 52.227-19 (Commercial Computer        *
*  Software - Restricted Rights) and DFAR 252.227-7013 (c) (1) (ii)       *
*  (Rights in Technical Data and Computer Software), as applicable.       *
***************************************************************************

   Entry Points: N/A 

   Modification History:

   Notes and restrictions on use:
       This module provides some generalised utilities for creating and
       managing a linked list.  Currently, only forward linked lists are
       managed by this code since they cover 80% of all cases.  It could
       easily be modified to handle doubly linked lists, but I'll leave
       that up to someone else to do.

       A special structure is used here which is defined by the typedef
       "llltype".  This structure allows the forward pointers to be
       generic across all platforms.  All the user must ensure is that
       the first entry in the structure to be managed by this link list
       utility be a forward pointer to the next node in the list.  To 
       guarantee portability across all platforms, the user could easily
       create a union of the llltype structure and the structure used in
       the calling program.

       There will be a need unders DOS to provide different versions of
       these libraries for each of the possible memory models.  This is,
       of course, a real bummer.  DOS can have two byte pointers or four
       byte pointers.  The code is written here to be able to handle both
       cases, but different compilations will effect the resultant library.
       The user will be responsible for determining whether or not the
       "far" pointers are required.
       
***************************************************************************/
/**************************************************************************/
/*  STANDARD C INCLUDE FILES  */
/**************************************************************************/

#include  <stdio.h>
#include  "sdk.h"


/**************************************************************************/
/*  GLOBAL DEFINES  */
/**************************************************************************/

#define LLL_STORE_END          'E'  /*  store type end  */
#define LLL_STORE_HEAD         'H'  /*  store type head  */


/**************************************************************************/
/*  GLOBAL TYPEDEFS  */
/**************************************************************************/

typedef struct lll {                /* LINKED LIST STRUCTURE */
                    struct lll *next; /*  forward pointer */
                    char *data;       /*  data for linked list */
                   } llltype;


/**************************************************************************/
/*  GLOBAL FUNCTIONS  */
/**************************************************************************/

char           *lll_alloc_node();   /*  node allocation function  */
int            lll_delete_node();   /*  node deletion function  */
int            lll_free_node();     /*  node free function  */
int            lll_store_node();    /*  node store function  */
char           *malloc();           /*  memory alloc function  */
void           free();              /*  memory free function  */


/**************************************************************************/
/*.doc lll_alloc_node(external)*/
/*+
   This function is used to allocate a node for storing in a linked list.
   The node is completely zeroed, except for the forward pointer which is
   made NULL.  Although on some systems this maybe zero, there is no
   guarantee that a NULL pointer is zero, and therefore, this is handled
   explicitly.  The function returns a char pointer which is the buffer
   that was allocated.  The user is responsible for casting the pointer to
   the correct type.  If NULL is returned, then the memory for the node
   could not be allocated.  The size argument is the size of the structure
   to be allocated.  This value is in bytes and must include the size of
   the forward pointer.
-*/
/**************************************************************************/
char
/* FCN */*lll_alloc_node(size)
  int size;                         /* (I) size in bytes of node */
{
    llltype *node;                  /* pointer to linked list structure */


    /* ALLOCATE THE NODE */
    if ((node = (llltype *)malloc(size)) == NULL)
        return NULL;
   

    /* INITIALIZE THE FORWARD POINTER FOR THE NODE */
    node->next = NULL;


    /* RETURN A POINTER TO THE NODE */
    return (char *)node;
}                                   /* END LLL_ALLOC_NODE FUNCTION */




/**************************************************************************/
/*.doc lll_delete_node(external)*/
/*+
   This function is used to delete a node from the linked list.  The linked
   list is updated to reflect the change, however, the user is responsible
   for calling lll_free_node to ensure that the memory used by the node is
   freed correctly.

   The function returns BAD if it couldn't locate the node in the specified
   linked list, and returns GOOD otherwise.
-*/
/**************************************************************************/
/* FCN */lll_delete_node(node,head)
  llltype *node;                    /* (I) node to delete */
  llltype **head;                   /* (I) head of linked list */
{
    llltype *n;                     /* local node pointer */


    /* DELETE THE SPECIFIED NODE */
    if (*head != NULL) {
        if (*head == node)
            *head = node->next;
        else {
            for ( n=*head; n->next!=node && n->next!=NULL; n=n->next);
                if (n->next == NULL)
                    return BAD;

                n->next = node->next;
        }
        return GOOD;
    }

    /* RETURN BAD STATUS TO CALLER */
    return BAD;
}                                   /* END LLL_DELETE_NODE FUNCTION */




/**************************************************************************/
/*.doc lll_free_node(external)*/
/*+
   This function is used to free a node which has been deleted from the
   linked list.  The user is responsible for deleting the node and then
   calling this function to delete the node.
-*/
/**************************************************************************/
/* FCN */lll_free_node(node)
    llltype *node;                  /* (I) node to be freed */
{
    /* FREE THE MEMORY ALLOCATED TO THE NODE */
    free((char *)node);

    /* RETURN GOOD STATUS TO CALLER */
    return GOOD;
}                                   /*  END LLL_FREE_NODE FUNCTION  */




/**************************************************************************/
/*.doc lll_store_node(external)*/
/*+
   This function is used to store a node into the linked list.  There are
   two ways a node can be stored in the linked list.  It can be stored at
   the head of the list as would be the case in a last in first out (LIFO)
   list, which also could be a stack.  It also can be stored at the end
   of the list as is the case in first in first out (FIFO) lists, which
   would be the queue case.  

   The function returns BAD if an invalid type was specified, and returns
   GOOD otherwise.
-*/
/**************************************************************************/
/* FCN */lll_store_node(node,head,type)
  llltype *node;                    /* (I) node to store */
  llltype **head;                   /* (I) head of linked list */
  char type;                        /* (I) type of store */
{
    llltype *n;                     /* local node pointer */


    /* STORE NODE IN LIST BASED ON THE SPECIFIED TYPE */
    switch (type) {
        case LLL_STORE_END:
            if (*head != NULL) {
                for (n=*head; n->next!=NULL; n=n->next);
                    n->next = node;
            } else
                *head = node;

            node->next = NULL;
            break;

        case LLL_STORE_HEAD:
            node->next = *head;
            *head = node;
            break;

        default:
            return BAD;
    }

    /* RETURN SUCCESS TO CALLER */
    return GOOD;
}                                   /* END LLL_STORE_NODE FUNCTION */
