/*
 * PROFILE.C
 *
 * MS-Windows-like private profile functions - see accompanying PROFILE.TXT
 * for documentation.
 *
 * Written May 1993 by Charlie Beerman.  This file may be distributed freely
 * as long as this header is maintained as part of the file.  If you make
 * changes, you must add a comment to this header indicating who you are
 * and what the changes are (so that I am not blamed for bugs that are not
 * mine!).
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "profile.h"

/* Local forward declarations */

static int deleteSection( char *sectionName );
static int deleteItem( char *sectionName, char *itemName );
static void deleteList( void );
static int readFile( char *filename );
static int writeFile( char *filename );
static int strcmpNoCase( char *s1, char *s2 );
static char *strTrimDequote( char *in, int doDequote );
static int atox( char *in );

/* Typedefs */

typedef enum LineType_ { Section, Item, Comment } LineTypeT;

typedef struct NodeStruct_ {

                   struct NodeStruct_ *nextNode;
                   LineTypeT lineType;
                   char *itemName;
                   char *itemValue;

} NodeStructT, *NodeStructPtrT;

/* Static Declarations */

static char *lastFilename = NULL;
static NodeStructPtrT firstNode = NULL;
static int currentListModified = 0;


/* WritePrivateProfileString
 *
 * Enters a profile string into the file
 *
 * Inputs:
 *
 *    char     *lpszSection     - Null terminated string containing the
 *                                section into which to place the item.
 *                                This should NOT include the square
 *                                brackets around the name.
 *
 *    char     *lpszEntry       - Null terminated string containing the
 *                                name of the item.  If this is NULL,
 *                                the section is deleted.
 *
 *    char     *lpszString      - Null terminated string containing the
 *                                value of the item.  If this is NULL,
 *                                the item specified by lpszEntry is
 *                                deleted.
 *
 *    char     *lpszFilename    - Null terminated string containing the
 *                                name of the file to be written to.
 *
 * Returns:
 *
 *    int                       - Nonzero for SUCCESS, zero for FAILURE
 *
 * Note:
 *
 *    The MS-Windows implementation caches the contents of the file
 *    but it seems to flush it after each write anyway, or at least does
 *    so when the Windows application is exited.  Since this implementation
 *    of the function does not run under the control of Windows, I have
 *    implemented it to cache the contents and write to disk only when
 *    a different file is specified or when all arguments except the
 *    filename are NULL.  In the latter case, the filename must be the
 *    file to be updated; it cannot be a different filename.
 */

int WritePrivateProfileString( char *lpszSection,
                               char *lpszEntry,
                               char *lpszString,
                               char *lpszFilename )
{
    NodeStructPtrT currentNode, newNode;
    NodeStructPtrT prevNode = NULL;

    /* Check for valid inputs */

    if ( lpszFilename == NULL )
        return (0);

    /*
     * See if we want to flush the cache.  We only do this if there have
     * been changes to the current list, and there is a current list,
     * and the filename of the current list is the same as the specified
     * filename.
     */

    if ( (lpszSection == NULL) && (lpszEntry == NULL) && (lpszString == NULL) ) {
        if ( (lastFilename != NULL) &&
             (strcmpNoCase( lastFilename, lpszFilename ) == 0) ) {
            if ( writeFile( lpszFilename ) == 0 ) {
                currentListModified = 0;
                deleteList();
                free( lastFilename );
                lastFilename = NULL;
                return (1);
            }
            else
                return (0);
        }
        else
            return (0);
    }

    /*
     * See if we want to flush the cache and read in a new list.
     */

     if ( (lastFilename == NULL) ||
             (strcmpNoCase( lastFilename, lpszFilename ) != 0) ) {

         /* Flush the cache if there was a modified current list */

         if ( currentListModified && (lastFilename != NULL) ) {
             if ( writeFile( lastFilename ) != 0 )
                 return (0);
         }

         deleteList();
         currentListModified = 0;
         if ( lastFilename != NULL )
             free( lastFilename );
         lastFilename = NULL;

         /* Read in the new list */

         readFile( lpszFilename );
         currentListModified = 0;
         lastFilename = malloc( strlen( lpszFilename ) + sizeof(char) );
         strcpy( lastFilename, lpszFilename );
     }

     /* See if we want to delete a section or item */

     if ( lpszEntry == NULL )
         return (deleteSection( lpszSection ));
     else if ( lpszString == NULL )
         return (deleteItem( lpszSection, lpszEntry ));

     /* If we get here, we want to add an entry */

     /* The task is to find the section */

     currentNode = firstNode;

     while ( currentNode != NULL ) {
        if ( currentNode->lineType == Section )
            if ( strcmpNoCase( currentNode->itemName, lpszSection ) == 0 )
                break;
        prevNode = currentNode;
        currentNode = currentNode->nextNode;
     }

     /* Do we have to create a section? */

     if ( currentNode == NULL ) {
         newNode = (NodeStructPtrT) calloc( 1, sizeof(NodeStructT) );
         newNode->lineType = Section;
         newNode->itemName = malloc( strlen( lpszSection ) + sizeof(char) );
         strcpy( newNode->itemName, lpszSection );

         if ( prevNode != NULL )
             prevNode->nextNode = newNode;
         else if ( firstNode == NULL )
             firstNode = newNode;
         else
             return (0);

         currentNode = newNode;
     }

     /* OK, we're on the section marker; now look for the item */

     while ( currentNode != NULL ) {
        if ( currentNode->lineType == Item )
            if ( strcmpNoCase( currentNode->itemName, lpszEntry ) == 0 )
                break;
        prevNode = currentNode;
        currentNode = currentNode->nextNode;

        /*
         * If next node is NULL or another section header,
         * create a new entry
         */

        if ( (currentNode == NULL) || (currentNode->lineType == Section) ) {
            newNode = (NodeStructPtrT) calloc( 1, sizeof(NodeStructT) );
            newNode->lineType = Item;
            newNode->itemName = malloc( strlen( lpszEntry ) + sizeof(char) );
            strcpy( newNode->itemName, lpszEntry );

            newNode->nextNode = currentNode;
            prevNode->nextNode = newNode;
            currentNode = newNode;
            break;
        }
     }

     /* At this point, the current node is the one we want to modify */

     if ( currentNode->itemValue != NULL )
         free( currentNode->itemValue );

     currentNode->itemValue = malloc( strlen( lpszString ) + sizeof(char) );
     strcpy( currentNode->itemValue, lpszString );

     currentListModified = 1;

     return (1);
}


/* GetPrivateProfileString
 *
 * Retrieves a string value from the profile
 *
 * Inputs:
 *
 *   char    *lpszSection      - Null terminated string containing the
 *                               profile section from which to retrieve
 *                               the value.
 *
 *   char    *lpszEntry        - Null terminated string containing the
 *                               name of the item from which to retrieve
 *                               the value.  If this is NULL, all item
 *                               names in the section are copied, separated
 *                               by nulls and terminated by two nulls.
 *
 *   char    *lpszDefault      - Null terminated string specifying the
 *                               default value if the specified entry
 *                               cannot be found.
 *
 *   char    *lpszReturnBuffer - Points to the buffer to receive the
 *                               return string.
 *
 *   int      cbReturnBuffer   - Size of the destination buffer.  The
 *                               destination buffer must be at least two
 *                               bytes long - one for a single data
 *                               character and one for the terminating null.
 *
 *   char    *lpszFilename     - Null terminated string containing the
 *                               name of the profile file.
 *
 * Returns:
 *
 *   int                       - Number of bytes copied to return buffer,
 *                               not including the terminating null.
 */

int GetPrivateProfileString( char *lpszSection,
                             char *lpszEntry,
                             char *lpszDefault,
                             char *lpszReturnBuffer,
                             int   cbReturnBuffer,
                             char *lpszFilename )
{
    NodeStructPtrT currentNode;
    char *valueString;
    char *returnPtr;
    int remainingLength;

    /* Initialize return buffer */

    if ( lpszReturnBuffer != NULL )
        *lpszReturnBuffer = '\0';

    /* See if inputs are valid (note that lpszEntry CAN be NULL) */

    if ( (lpszSection == NULL) || (lpszDefault == NULL) ||
         (lpszReturnBuffer == NULL) || (cbReturnBuffer < 2) ||
         (lpszFilename == NULL) )
        return (0);

    /*
     * See if we want to flush the cache and read in a new list.
     */

    if ( (lastFilename == NULL) ||
           (strcmpNoCase( lastFilename, lpszFilename ) != 0) ) {

        /* Flush the cache if there was a modified current list */

        if ( currentListModified && (lastFilename != NULL) ) {
            if ( writeFile( lastFilename ) != 0 )
                return (0);
        }

        deleteList();
        currentListModified = 0;
        if ( lastFilename != NULL )
            free( lastFilename );
        lastFilename = NULL;

        /* Read in the new list */

        readFile( lpszFilename );
        currentListModified = 0;
        lastFilename = malloc( strlen( lpszFilename ) + sizeof(char) );
        strcpy( lastFilename, lpszFilename );
    }

    /* Look for the desired section */

    currentNode = firstNode;
    while ( currentNode != NULL ) {
        if ( currentNode->lineType == Section )
            if ( strcmpNoCase( currentNode->itemName, lpszSection ) == 0 )
                break;
        currentNode = currentNode->nextNode;
    }

    /* If NULL, we couldn't find the section, so return the default */

    if ( currentNode == NULL ) {
        strncpy( lpszReturnBuffer, lpszDefault, cbReturnBuffer );
        if ( lpszEntry != NULL ) {
            lpszReturnBuffer[cbReturnBuffer-1] = '\0';
        }
        else {
            lpszReturnBuffer[cbReturnBuffer-1] = '\0';
            lpszReturnBuffer[cbReturnBuffer-2] = '\0';
        }
        return (strlen(lpszReturnBuffer));
    }

    /* Look for the desired item */

    returnPtr = lpszReturnBuffer;
    *lpszReturnBuffer = '\0';
    *(lpszReturnBuffer+1) = '\0';
    remainingLength = cbReturnBuffer - 2;

    while ( currentNode != NULL ) {

        /* Find the item */

        if ( currentNode->lineType == Item ) {

            /*
             * If lpszEntry is NULL, we want to copy each item name
             */

            if ( lpszEntry == NULL ) {
                strncpy( returnPtr, currentNode->itemName, remainingLength );
                returnPtr[remainingLength] = '\0';
                remainingLength -= strlen( currentNode->itemName );
                returnPtr += strlen( returnPtr );
                *returnPtr = '\0';
                remainingLength--;
                returnPtr++;
                if ( remainingLength <= 0 )
                    break;
            }

            /*
             * Otherwise, if the current entry is the one we're looking for,
             * return its string value.
             */

            else if ( strcmpNoCase( currentNode->itemName, lpszEntry ) == 0 ) {
                valueString = malloc( strlen( currentNode->itemValue ) + sizeof(char) );
                strcpy( valueString, currentNode->itemValue );
                strTrimDequote( valueString, 1 );
                strncpy( lpszReturnBuffer, valueString, cbReturnBuffer );
                lpszReturnBuffer[cbReturnBuffer-1] = '\0';
                free( valueString );
                return (strlen( lpszReturnBuffer ));
            }
        }

        /* Look for next node */

        currentNode = currentNode->nextNode;

        /* If next node is NULL or another section header, quit loop */

        if ( currentNode != NULL )
            if ( currentNode->lineType == Section )
                break;
    }

    /*
     * If lpszEntry was not NULL, if we get here we could not find the item.
     * If lpszEntry was not NULL and we did find the item, we should never get here.
     * If lpszEntry was NULL, when we get here we should just return whatever
     * we did find, after adding a second terminating NULL.
     */

    if ( (currentNode == NULL) && (lpszEntry != NULL) ) {
        strncpy( lpszReturnBuffer, lpszDefault, cbReturnBuffer );
        lpszReturnBuffer[cbReturnBuffer-1] = '\0';
        return (strlen( lpszReturnBuffer ));
    }
    else {
        *returnPtr = '\0';
        return (returnPtr - lpszReturnBuffer - 1);
    }
}


/* GetPrivateProfileInt
 *
 * Retrieves an integer from the profile
 *
 * Inputs:
 *
 *   char    *lpszSection      - Null terminated string containing the
 *                               profile section from which to retrieve
 *                               the integer.
 *
 *   char    *lpszEntry        - Null terminated string containing the
 *                               name of the item from which to retrieve
 *                               the integer.
 *
 *   int      deflt            - The default integer to use if the entry
 *                               is not found.
 *
 *   char    *lpszFilename     - Null terminated string containing the
 *                               name of the profile file.
 *
 * Returns:
 *
 *   unsigned int              - The resulting integer.  If you want a
 *                               signed integer, cast the return value
 *                               to a signed integer.
 *
 */

unsigned int GetPrivateProfileInt( char *lpszSection,
                                   char *lpszEntry,
                                   int   deflt,
                                   char *lpszFilename )
{
    NodeStructPtrT currentNode;
    char *valueString;
    int value;

    /* See if inputs are valid */

    if ( (lpszSection == NULL) || (lpszEntry == NULL) || (lpszFilename == NULL) )
        return ((unsigned int)deflt);

    /*
     * See if we want to flush the cache and read in a new list.
     */

    if ( (lastFilename == NULL) ||
           (strcmpNoCase( lastFilename, lpszFilename ) != 0) ) {

        /* Flush the cache if there was a modified current list */

        if ( currentListModified && (lastFilename != NULL) ) {
            if ( writeFile( lastFilename ) != 0 )
                return (0);
        }

        deleteList();
        currentListModified = 0;
        if ( lastFilename != NULL )
            free( lastFilename );
        lastFilename = NULL;

        /* Read in the new list */

        readFile( lpszFilename );
        currentListModified = 0;
        lastFilename = malloc( strlen( lpszFilename ) + sizeof(char) );
        strcpy( lastFilename, lpszFilename );
    }

    /* Look for the desired section */

    currentNode = firstNode;
    while ( currentNode != NULL ) {
        if ( currentNode->lineType == Section )
            if ( strcmpNoCase( currentNode->itemName, lpszSection ) == 0 )
                break;
        currentNode = currentNode->nextNode;
    }

    /* If NULL, we couldn't find the section, so return the default */

    if ( currentNode == NULL )
        return ((unsigned int)deflt);

    /* Look for the desired item */

    while ( currentNode != NULL ) {

        /* Find the item */

        if ( currentNode->lineType == Item )
            if ( strcmpNoCase( currentNode->itemName, lpszEntry ) == 0 )
                break;
        currentNode = currentNode->nextNode;

        /* If next node is NULL or another section header, quit loop */

        if ( currentNode != NULL )
            if ( currentNode->lineType == Section )
                break;
    }

    /* If NULL, we couldn't find the item, so return the default */

    if ( currentNode == NULL )
        return ((unsigned int)deflt);

    /* Convert the item string to an integer */

    if ( currentNode->itemValue == NULL )
        return ((unsigned int)deflt);

    valueString = malloc( strlen( currentNode->itemValue ) + sizeof(char) );
    strcpy( valueString, currentNode->itemValue );
    strTrimDequote( valueString, 1 );

    value = atox( valueString );

    free( valueString );
    return ((unsigned int)value);
}


static int deleteSection( char *sectionName )
{
    NodeStructPtrT currentNode, prevNode;

    /* Check for valid section name */

    if ( sectionName == NULL )
        return (0);

    /* Initialize */

    currentNode = firstNode;
    prevNode = NULL;

    /* Loop through list, looking for desired section */

    while ( currentNode != NULL ) {
        if ( currentNode->lineType == Section )
            if ( strcmpNoCase( currentNode->itemName, sectionName ) == 0 )
                break;
        prevNode = currentNode;
        currentNode = currentNode->nextNode;
    }

    /* If NULL, we couldn't find the section, so return an error status */

    if ( currentNode == NULL )
        return (0);

    /* Loop, beginning with section header */

    while ( currentNode != NULL ) {

        /* Update the linked list */

        if ( prevNode == NULL )
            firstNode = currentNode->nextNode;
        else
            prevNode->nextNode = currentNode->nextNode;

        /* Delete the current node */

        if ( currentNode->itemName != NULL )
            free( currentNode->itemName );
        if ( currentNode->itemValue != NULL )
            free( currentNode->itemValue );
        free( currentNode );

        currentListModified = 1;

        /* Get the next node */

        if ( prevNode == NULL )
            currentNode = firstNode;
        else
            currentNode = prevNode->nextNode;

        /* If next node is NULL or another section header, quit loop */

        if ( currentNode != NULL )
            if ( currentNode->lineType == Section )
                break;
    }

    return (1);
}


static int deleteItem( char *sectionName, char *itemName )
{
    NodeStructPtrT currentNode, prevNode;

    /* Check for valid section and item name */

    if ( sectionName == NULL )
        return (0);

    if ( itemName == NULL )
        return (deleteSection( sectionName ));

    /* Initialize */

    currentNode = firstNode;
    prevNode = NULL;

    /* Loop through list, looking for desired section */

    while ( currentNode != NULL ) {
        if ( currentNode->lineType == Section )
            if ( strcmpNoCase( currentNode->itemName, sectionName ) == 0 )
                break;
        prevNode = currentNode;
        currentNode = currentNode->nextNode;
    }

    /* If NULL, we couldn't find the section, so return an error status */

    if ( currentNode == NULL )
        return (0);

    /* Loop, beginning with section header */

    while ( currentNode != NULL ) {

        /* Find the item */

        if ( currentNode->lineType == Item )
            if ( strcmpNoCase( currentNode->itemName, itemName ) == 0 ) {

                /* Update the linked list */

                if ( prevNode == NULL )
                    firstNode = currentNode->nextNode;
                else
                    prevNode->nextNode = currentNode->nextNode;

                /* Delete the current node */

                if ( currentNode->itemName != NULL )
                    free( currentNode->itemName );
                if ( currentNode->itemValue != NULL )
                    free( currentNode->itemValue );
                free( currentNode );

                currentListModified = 1;

                /* We're done */

                return (1);
            }

        /* Get the next node */

        prevNode = currentNode;
        currentNode = currentNode->nextNode;

        /* If next node is NULL or another section header, quit loop */

        if ( currentNode != NULL )
            if ( currentNode->lineType == Section )
                break;
    }

    /* If we get here, we couldn't find the item */

    return (0);
}


/* readFile
 *
 * Reads an initialization file to the list.
 *
 * Inputs:  char *filename        - File to read
 *
 * Returns: int  =  0             - Success
 *
 *               != 0             - Failure
 */

static int readFile( char *filename )
{
    FILE *fp;
    char *buffer;
    char *eqLoc;
    char *startLoc;
    char *closeLoc;
    NodeStructPtrT node;
    NodeStructPtrT lastNode = NULL;

    /* Open the file */

    fp = fopen( filename, "r" );
    if ( fp == NULL )
        return (-1);

    /* Create a large buffer to hold the lines of the file as we read them */

    buffer = (char *) calloc( 20480, sizeof(char) );

    /* Loop over the lines of the file, processing each one in turn */

    while ( !feof( fp ) ) {
        if ( fgets( buffer, 20480, fp ) == NULL )
            break;

        /* Create a new node */
                            
        node = (NodeStructPtrT) calloc( 1, sizeof(NodeStructT) );

        /* Figure out what kind of item it is */

        startLoc = buffer + strspn( buffer, " \t\n" );

        if ( *startLoc == ';' ) {                  /* Comment */
            node->lineType = Comment;
            node->itemName = malloc( strlen( startLoc ) + sizeof(char) );
            strcpy( node->itemName, startLoc );
            strTrimDequote( node->itemName, 0 );
        }

        else if ( *startLoc == '[' ) {             /* Section */
            node->lineType = Section;
            node->itemName = malloc( strlen( startLoc ) + sizeof(char) );
            strcpy( node->itemName, startLoc + 1 );
            closeLoc = strchr( node->itemName, ']' );
            if ( closeLoc != NULL )
                *closeLoc = '\0';

            strTrimDequote( node->itemName, 1 );
        }

        else {                                     /* Item */
            node->lineType = Item;
            eqLoc = strchr( startLoc, '=' );
            if ( eqLoc == NULL ) {
                free( node );
                continue;
            }

            *eqLoc = '\0';
            node->itemName = malloc( strlen( startLoc ) + sizeof(char) );
            strcpy( node->itemName, startLoc );
            strTrimDequote( node->itemName, 1 );

            eqLoc++;
            node->itemValue = malloc( strlen( eqLoc ) + sizeof(char) );
            strcpy( node->itemValue, eqLoc );
            strTrimDequote( node->itemValue, 0 );
        }

        /* Add the node to the linked list */

        if ( firstNode == NULL )
            firstNode = node;
        else
            lastNode->nextNode = node;

        lastNode = node;
    }

    /* Free the buffer, close the file and return */

    free( buffer );
    fclose( fp );
    return (0);
}


/* writeFile
 *
 * Writes an initialization file from the list.
 *
 * Inputs:  char *filename        - File to write
 *
 * Returns: int  =  0             - Success
 *
 *               != 0             - Failure
 */

static int writeFile( char *filename )
{
    FILE *fp;
    NodeStructPtrT currentNode;

    /* Open the file */

    fp = fopen( filename, "w" );
    if ( fp == NULL )
        return (-1);

    /* Loop over the list, writing each line */

    currentNode = firstNode;
    while ( currentNode != NULL ) {

        /* Comments get output as is, since they were read as is */

        if ( currentNode->lineType == Comment ) {
            fprintf( fp, "%s\n", currentNode->itemName );
        }

        /*
         * SECTIONS get put in square brackets, with a blank line before
         * all but the first.
         */

        else if ( currentNode->lineType == Section ) {
            if ( currentNode != firstNode )
                fprintf( fp, "\n" );
            if ( currentNode->itemName != NULL )
                fprintf( fp, "[%s]\n", currentNode->itemName );
        }

        /*
         * ITEMS get a line of the form ITEM=VALUE
         */

        else if ( currentNode->lineType == Item ) {
            if ( (currentNode->itemName != NULL) && (currentNode->itemValue != NULL) )
                fprintf( fp, "%s=%s\n",
                         currentNode->itemName, currentNode->itemValue );
        }

        currentNode = currentNode->nextNode;
    }

    return (fclose( fp ));
}


/* Function to delete the current list from memory */

static void deleteList( void )
{
    NodeStructPtrT currentNode, tmp;

    currentNode = firstNode;

    while ( currentNode != NULL ) {
        tmp = currentNode;
        if ( currentNode->itemName != NULL )
            free( currentNode->itemName );
        if ( currentNode->itemValue != NULL )
            free( currentNode->itemValue );

        currentNode = currentNode->nextNode;
        free( tmp );
    }

    firstNode = NULL;
}


/* Case Insensitive String Compare */

static int strcmpNoCase( char *s1, char *s2 )
{
    /* Check for NULL inputs */

    if ( s1 == NULL )
        if ( s2 == NULL )
            return (0);
        else
            return (-1);
    else if ( s2 == NULL )
        return (1);

    while ( (*s1 != '\0') && (*s2 != '\0') ) {

        if ( toupper( *s1 ) < toupper( *s2 ) )
            return (-1);
        else if ( toupper( *s1 ) > toupper( *s2 ) )
            return (1);
        else {
            s1++;
            s2++;
        }

    }

    if ( (*s1 == '\0') && (*s2 == '\0') )
        return (0);
    else if ( *s1 == '\0' )
        return (-1);
    else
        return (1);
}


/* String trim and dequote
 *
 * Trims leading and trailing whitespace and removes quotes.
 *
 * Returns its input with whitespace and quotes removed.
 */

static char *strTrimDequote( char *in, int doDequote )
{
    char *startPos;
    char *endPos;
    char *tmp;

    /* Get past any whitespace and past the initial quote */

    startPos = in;
    while ( *startPos != '\0' ) {
        if ( (*startPos == ' ') || (*startPos == '\t') || (*startPos == '\n') ) {
            startPos++;
            continue;
        }

        else if ( doDequote &&
                  ((*startPos == '\"') || (*startPos == '\'')) ) {
            startPos++;
            break;
        }

        else
            break;
    }

    /* If there's nothing left, just truncate the input and return */

    if ( *startPos == '\0' ) {
        *in = '\0';
        return (in);
    }

    /* Copy the remaining string to a temporary buffer */

    tmp = malloc( strlen( startPos ) + sizeof(char) );
    strcpy( tmp, startPos );

    /* Now go from the end of the buffer and look for last valid char. */

    endPos = tmp + strlen( tmp ) - 1;
    while ( endPos >= tmp ) {
        if ( (*endPos == ' ') || (*endPos == '\t') || (*endPos == '\n' ) ) {
            endPos--;
            continue;
        }

        else if ( doDequote &&
                  ((*endPos == '\"') || (*endPos == '\'')) ) {
            endPos--;
            break;
        }

        else
            break;
    }

    *(endPos + 1) = '\0';

    /* Now copy back to the input string and free temporary buffer */

    *in = '\0';
    strcpy( in, tmp );

    free( tmp );

    return (in);
}


/*
 * atox
 *
 * Converts hex, octal, or decimal string to an integer
 *
 * Inputs:
 *
 *   char *in   - Input string
 *
 * Returns:
 *
 *   int        - Signed integer value of string
 *
 * Notes:
 *
 *   Hexadecimal values are preceded by an 0x, 0X, -0x, or -0X.
 *   Octal values are preceded by a 0 or -0.
 *
 *   Whitespace is not allowed inside the value.
 */

static int atox( char *in )
{
    int value;
    int signMultiplier;
    int radix;
    char *cp;

    static char digits[] = "0123456789ABCDEF";

    /* Skip any leading whitespace */

    cp = in + strspn( in, " \t\n" );

    if ( *cp == '-' ) {
        cp++;
        signMultiplier = -1;
    }
    else
        signMultiplier = 1;

    /* Figure out if this is hex or octal or decimal */

    radix = 10;
    if ( *cp == '0' ) {
        cp++;
        if ( (*cp == 'x') || (*cp == 'X') ) {
            cp++;
            radix = 16;
        }
        else
            radix = 8;
    }

    /* Loop over digits in the string */

    value = 0;

    while ( *cp != '\0' ) {

        /* Check for valid character for radix */

        switch (radix) {

            case 8:
                if ( (*cp < '0') || (*cp > '7') )
                    return (value * signMultiplier);
                break;

            case 10:
                if ( (*cp < '0') || (*cp > '9') )
                    return (value * signMultiplier);
                break;

            case 16:
                if ( strchr( digits, toupper(*cp) ) == NULL )
                    return (value * signMultiplier);
                break;

            default:
                return 0;
        }

        value = (value * radix) + (strchr( digits, toupper(*cp) ) - digits);

        cp++;
    }

    return (value * signMultiplier);
}
